import importlib
from copy import deepcopy
from typing import Dict, Union, Optional, List

import numpy as np
import pandas as pd
import mxnet as mx
from sklearn.preprocessing import OrdinalEncoder

from gluonts.dataset.pandas import PandasDataset

from gluonts.mx import DeepAREstimator, Trainer  # @TODO: support for other estimators
from gluonts.mx.trainer.callback import TrainingHistory
from gluonts.mx.distribution.student_t import StudentTOutput

from lightwood.helpers.log import log
from lightwood.mixer.base import BaseMixer
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs


class GluonTSMixer(BaseMixer):
    horizon: int
    target: str
    supports_proba: bool

    def __init__(
            self,
            stop_after: float,
            target: str,
            horizon: int,
            window: int,
            dtype_dict: Dict,
            ts_analysis: Dict,
            n_epochs: int = 10,
            early_stop_patience: int = 3,
            distribution_output: str = '',
            seed: int = 0,
            static_features_cat: Optional[List[str]] = None,
            static_features_real: Optional[List[str]] = None,
    ):
        """
        Wrapper around GluonTS probabilistic deep learning models. For now, only DeepAR is supported.

        Due to inference speed, predictions are only generated for the last data point (as opposed to other mixers).  

        :param stop_after: time budget in seconds.
        :param target: column to forecast.
        :param horizon: length of forecasted horizon.
        :param window: length of input data.
        :param dtype_dict: data type of each column in the dataset.
        :param ts_analysis: dictionary with miscellaneous time series info, as generated by 'lightwood.data.timeseries_analyzer'.
        :param n_epochs: amount of epochs to train the model for. Will perform early stopping automatically if validation loss degrades.
        :param early_stop_patience: amount of consecutive epochs with no improvement in the validation loss.
        :param distribution_output: specify the type of distribution the model will learn.
        :param seed: specifies the seed used internally by GluonTS for reproducible predictions
        """  # noqa
        super().__init__(stop_after)
        self.stable = True
        self.prepared = False
        self.supports_proba = False
        self.target = target
        self.window = window
        self.horizon = horizon
        self.n_epochs = n_epochs
        self.dtype_dict = dtype_dict
        self.ts_analysis = ts_analysis
        self.grouped_by = ['__default'] if not ts_analysis['tss'].group_by else ts_analysis['tss'].group_by
        self.groups = []  # list can grow using adjust() with new data
        self.estimator = None
        self.model = None
        self.train_cache = None
        self.patience = early_stop_patience
        self.seed = seed
        self.trains_once = True
        self.static_features_cat_encoders = {}
        self.static_features_cat = static_features_cat if static_features_cat else []
        self.static_features_real = static_features_real if static_features_real else []

        dist_module = importlib.import_module('.'.join(['gluonts.mx.distribution',
                                                        *distribution_output.split(".")[:-1]]))
        try:
            self.distribution = getattr(dist_module, distribution_output.split(".")[-1])()
        except AttributeError:
            self.distribution = StudentTOutput()  # use StudentTOutput when the provided distribution does not exist

        if len(self.grouped_by) > 1:
            raise Exception("This mixer can only be used with 0 or 1 partition columns.")

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """ Fits the model. """  # noqa
        log.info('Started fitting GluonTS forecasting model')

        # prepare data
        cat_ds = ConcatedEncodedDs([train_data, dev_data])
        for col in self.static_features_cat:
            self.static_features_cat_encoders[col] = OrdinalEncoder().fit(cat_ds.data_frame[col].values.reshape(-1, 1))

        fit_groups = list(cat_ds.data_frame[self.grouped_by[0]].unique()) if self.grouped_by != ['__default'] else None
        train_ds = self._make_initial_ds(cat_ds.data_frame, phase='train', groups=fit_groups)
        batch_size = 32
        self.model_train_stats = TrainingHistory()

        self.estimator = DeepAREstimator(
            freq=train_ds.freq,
            prediction_length=self.horizon,
            distr_output=self.distribution,
            lags_seq=[i + 1 for i in range(self.window)],
            batch_size=batch_size,
            use_feat_static_cat=True if self.static_features_cat else False,
            use_feat_static_real=True if self.static_features_real else False,
            trainer=Trainer(
                epochs=self.n_epochs,
                num_batches_per_epoch=max(1, len(cat_ds.data_frame) // batch_size),
                callbacks=[EarlyStop(patience=self.patience), self.model_train_stats])
        )
        self.model = self.estimator.train(train_ds)
        self.prepared = True
        log.info('Successfully trained GluonTS forecasting model.')

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs,
                    args: Optional[dict] = None) -> None:
        """
        Due to how lightwood implements the `update` procedure, expected inputs for this method are:

        :param dev_data: original `test` split (used to validate and select model if ensemble is `BestOf`).
        :param train_data: concatenated original `train` and `dev` splits.
        :param args: optional arguments to customize the finetuning process.
        """  # noqa

        # handle args
        if args is None:
            args = {}
        if args.get('n_epochs'):
            self.n_epochs = args['n_epochs']
        if args.get('patience'):
            self.patience = args['patience']
        self.estimator.trainer = Trainer(epochs=self.n_epochs, callbacks=[EarlyStop(patience=self.patience)])

        # prepare data and fine-tune
        ds = ConcatedEncodedDs([train_data, dev_data])
        adjust_groups = list(ds.data_frame[self.grouped_by[0]].unique()) if self.grouped_by != ['__default'] else None
        ds = self._make_initial_ds(ds.data_frame, phase='adjust', groups=adjust_groups)
        self.model = self.estimator.train_from(self.model, ds)

    def __call__(self, ds: Union[EncodedDs, ConcatedEncodedDs],
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        """ 
        Calls the mixer to emit forecasts.
        """  # noqa
        mx.random.seed(self.seed)
        np.random.seed(self.seed)
        length = sum(ds.encoded_ds_lengths) if isinstance(ds, ConcatedEncodedDs) else len(ds)

        ydf = pd.DataFrame(index=np.arange(length), dtype=object)
        init_arr = [0 for _ in range(self.ts_analysis['tss'].horizon)]
        for col in ['prediction', 'lower', 'upper']:
            ydf.at[:, col] = [init_arr for _ in range(len(ydf))]

        ydf['index'] = ds.data_frame.index
        conf = args.fixed_confidence if args.fixed_confidence else 0.9
        ydf['confidence'] = conf

        gby = self.ts_analysis["tss"].group_by if self.ts_analysis["tss"].group_by else []
        groups = ds.data_frame[gby[0]].unique().tolist() if gby else None

        df = ds.data_frame
        ydf['__original_index'] = df['__mdb_original_index'].values
        input_ds = self._make_initial_ds(df, groups=groups)  # TODO test with novel group
        forecasts = list(self.model.predict(input_ds))
        grouped = df.groupby(by=gby) if gby else df.groupby(lambda x: '__default')
        for (group, subdf), group_forecast in zip(grouped, forecasts):
            idx = ydf[ydf['__original_index'] == max(subdf['__mdb_original_index'])].index.values[0]
            ydf.at[idx, 'prediction'] = [entry for entry in group_forecast.quantile(0.5)]
            ydf.at[idx, 'lower'] = [entry for entry in group_forecast.quantile(1 - conf)]
            ydf.at[idx, 'upper'] = [entry for entry in group_forecast.quantile(conf)]

        return ydf

    def _make_initial_ds(self, df=None, phase='predict', groups=None):
        oby_col_name = '__gluon_timestamp'
        gby = self.ts_analysis["tss"].group_by if self.ts_analysis["tss"].group_by else []
        freq = self.ts_analysis['sample_freqs']['__default']
        keep_cols = [self.target] + [col for col in gby] + self.static_features_cat + self.static_features_real

        agg_map = {self.target: 'sum'}
        for col in self.static_features_cat:
            agg_map[col] = 'first'
        for col in self.static_features_real:
            agg_map[col] = 'mean'

        if groups is None and gby:
            groups = self.groups
        elif gby and phase in ('train', 'adjust') and self.grouped_by != ['__default']:
            # we extend all seen groups for subsequent adjustments
            self.groups.extend(set(groups))

        if df is None and phase not in ('train', 'adjust'):
            df = self.train_cache
            if gby:
                df = df[df[gby[0]].isin(groups)]
        else:
            sub_df = df[keep_cols]
            df = deepcopy(sub_df)

            if phase == 'train':
                self.train_cache = df.sort_index()
            else:
                if gby:
                    cache = self.train_cache[self.train_cache[gby[0]].isin(groups)]
                else:
                    cache = self.train_cache

                if phase == 'adjust':
                    # update cache to include all new information (pre-group filter)
                    self.train_cache = pd.concat([self.train_cache, df]).drop_duplicates().sort_index()

                df = pd.concat([cache, df]).sort_index()

        df = df.drop_duplicates()

        if len(df) == 0:
            return None

        if gby:
            # @TODO: multiple group support and remove groups without enough data
            df = df.groupby(by=gby[0]).resample(freq).agg(agg_map).reset_index(level=[0])
        else:
            df = df.resample(freq).agg(agg_map)
            gby = '__default_group'
            df[gby] = '__default_group'

        df[oby_col_name] = df.index
        for col in self.static_features_cat:
            df[col] = self.static_features_cat_encoders[col].transform(df[col].values.reshape(-1, 1))

        static_features = []
        if self.static_features_real:
            static_features.extend(self.static_features_real)
        if self.static_features_cat:
            static_features.extend(self.static_features_cat)
        static_features_df = df[static_features]

        ds = PandasDataset.from_long_dataframe(
            df,
            target=self.target,
            item_id=gby[0],
            freq=freq,
            timestamp=oby_col_name,
            static_features=static_features_df
        )
        return ds


class EarlyStop(TrainingHistory):
    def __init__(self, patience=3):
        super().__init__()
        self.patience = max(1, patience)
        self.counter = 0

    def on_validation_epoch_end(
        self,
        epoch_no: int,
        epoch_loss: float,
        training_network,
        trainer,
    ) -> bool:
        super().on_validation_epoch_end(epoch_no, epoch_loss, training_network, trainer)

        if len(self.validation_loss_history) > 1:
            if self.validation_loss_history[-1] > self.validation_loss_history[-2]:
                self.counter += 1
            else:
                self.counter = 0  # reset if not successive

        if self.counter >= self.patience:
            return False
        else:
            return True



# Encoders which should always work
from lightwood.encoder.base import BaseEncoder
from lightwood.encoder.datetime.datetime import DatetimeEncoder
from lightwood.encoder.datetime.datetime_sin_normalizer import DatetimeNormalizerEncoder
from lightwood.encoder.image.img_2_vec import Img2VecEncoder
from lightwood.encoder.numeric.numeric import NumericEncoder
from lightwood.encoder.numeric.ts_numeric import TsNumericEncoder
from lightwood.encoder.array.ts_num_array import TsArrayNumericEncoder
from lightwood.encoder.text.short import ShortTextEncoder
from lightwood.encoder.text.vocab import VocabularyEncoder
from lightwood.encoder.categorical.simple_label import SimpleLabelEncoder
from lightwood.encoder.categorical.onehot import OneHotEncoder
from lightwood.encoder.categorical.binary import BinaryEncoder
from lightwood.encoder.categorical.autoencoder import CategoricalAutoEncoder
from lightwood.encoder.time_series.ts import TimeSeriesEncoder
from lightwood.encoder.array.array import ArrayEncoder, NumArrayEncoder, CatArrayEncoder
from lightwood.encoder.categorical.multihot import MultiHotEncoder
from lightwood.encoder.array.ts_cat_array import TsCatArrayEncoder
from lightwood.encoder.text.pretrained import PretrainedLangEncoder
from lightwood.encoder.audio import MFCCEncoder


__all__ = ['BaseEncoder', 'DatetimeEncoder', 'Img2VecEncoder', 'NumericEncoder', 'TsNumericEncoder',
           'TsArrayNumericEncoder', 'ShortTextEncoder', 'VocabularyEncoder', 'OneHotEncoder',
           'CategoricalAutoEncoder', 'TimeSeriesEncoder', 'ArrayEncoder', 'MultiHotEncoder', 'TsCatArrayEncoder',
           'NumArrayEncoder', 'CatArrayEncoder', 'SimpleLabelEncoder',
           'PretrainedLangEncoder', 'BinaryEncoder', 'DatetimeNormalizerEncoder', 'MFCCEncoder']



import torch
import os
from random import randint
from torch.cuda import device_count, get_device_capability


def is_cuda_compatible():
    """
    Check if the system has CUDA-compatible devices with the required architecture and
    compiled CUDA version.

    This function checks the compatibility of CUDA devices available on the system by
    comparing their architectures and the compiled CUDA version. It iterates through
    the available devices and verifies if their architectures meet the minimum
    requirement specified by the function, and also checks if the compiled CUDA
    version is greater than a specific version.

    Returns:
        bool: True if there are compatible CUDA devices, otherwise False.

    Example:
        >>> is_compatible = is_cuda_compatible()
        >>> print(is_compatible)
        True
    """
    compatible_device_count = 0
    if torch.version.cuda is not None:
        for d in range(device_count()):
            capability = get_device_capability(d)
            major = capability[0]
            minor = capability[1]
            current_arch = major * 10 + minor
            min_arch = min((int(arch.split("_")[1]) for arch in torch.cuda.get_arch_list()), default=35)
            if (not current_arch < min_arch
                    and not torch._C._cuda_getCompiledVersion() <= 9000):
                compatible_device_count += 1

    if compatible_device_count > 0:
        return True
    return False


def get_devices():
    """
    Get the appropriate Torch device(s) based on CUDA availability and compatibility.

    This function determines the appropriate Torch device(s) to be used for
    computations based on the availability of CUDA and compatible devices. It checks
    if CUDA is available and if the available CUDA devices are compatible according to
    the 'is_cuda_compatible()' function. If compatible devices are found, the function
    selects either the first available CUDA device or a randomly selected one based on
    the 'RANDOM_GPU' environment variable. If CUDA is not available or no compatible
    devices are found, the function returns the CPU device.

    Returns:
        Tuple: A tuple containing the selected Torch device and the number of available
        devices.
    Example:
        >>> device, num_devices = get_devices()
        >>> print(device)
        cuda:0
        >>> print(num_devices)
        1
    """
    if torch.cuda.is_available() and is_cuda_compatible():
        device_str = "cuda"
        available_devices = torch.cuda.device_count()

        if available_devices > 1:
            if os.environ.get('RANDOM_GPU', False) in ['1', 'true', 'True', True, 1]:
                device_str = 'cuda:' + str(randint(0, available_devices - 1))
                available_devices = 1
    else:
        device_str = "cpu"
        available_devices = 0

    return torch.device(device_str), available_devices


def get_device_from_name(device_name=''):
    """
    Get a Torch device based on the specified device name or default behavior.

    This function returns a Torch device based on the specified device name or the
    default behavior, which is to return the output of the 'get_devices()' function.

    Args:
        device_name (str, optional): Name of the device to use. Default is an empty
        string.
        
    Returns:
        torch.device: The selected Torch device.

    Example:
        >>> device = get_device_from_name('cuda:1')
        >>> print(device)
        cuda:1
    """ # noqa E501
    if(device_name != ''):
        device = torch.device(device_name)
    else:
        device, _ = get_devices()
    return device



# Original author: Henrik Linusson (github.com/donlnz)
import abc
from typing import Dict
import numpy as np
from sklearn.base import BaseEstimator


from lightwood.analysis.nc.util import t_softmax


class RegressorMixin(object):
    def __init__(self) -> None:
        super(RegressorMixin, self).__init__()

    @classmethod
    def get_problem_type(cls):
        return 'regression'


class ClassifierMixin(object):
    def __init__(self) -> None:
        super(ClassifierMixin, self).__init__()

    @classmethod
    def get_problem_type(cls) -> str:
        return 'classification'


class TSMixin(object):
    def __init__(self) -> None:
        super(TSMixin, self).__init__()

    @classmethod
    def get_problem_type(cls):
        return 'time-series'


class BaseModelAdapter(BaseEstimator):
    __metaclass__ = abc.ABCMeta

    def __init__(self, model: object, fit_params: Dict[str, object] = None) -> None:
        super(BaseModelAdapter, self).__init__()

        self.model = model
        self.last_x, self.last_y = None, None
        self.clean = False
        self.fit_params = {} if fit_params is None else fit_params

    def fit(self, x: np.array, y: np.array) -> None:
        """Fits the model.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of examples for fitting the model.

        y : numpy array of shape [n_samples]
            Outputs of examples for fitting the model.

        Returns
        -------
        None
        """

        self.model.fit(x, y, **self.fit_params)
        self.clean = False

    def predict(self, x: np.array) -> np.array:
        """Returns the prediction made by the underlying model.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of test examples.

        Returns
        -------
        y : numpy array of shape [n_samples]
            Predicted outputs of test examples.
        """
        if (
                not self.clean or
                self.last_x is None or
                self.last_y is None or
                not np.array_equal(self.last_x, x)
        ):
            self.last_x = x
            self.last_y = self._underlying_predict(x)
            self.clean = True

        return self.last_y.copy()

    @abc.abstractmethod
    def _underlying_predict(self, x: np.array) -> np.array:
        """Produces a prediction using the encapsulated model.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of test examples.

        Returns
        -------
        y : numpy array of shape [n_samples]
            Predicted outputs of test examples.
        """
        pass


class ClassifierAdapter(BaseModelAdapter):
    def __init__(self, model: object, fit_params: Dict[str, object] = None) -> None:
        super(ClassifierAdapter, self).__init__(model, fit_params)

    def _underlying_predict(self, x: np.array) -> np.array:
        return self.model.predict_proba(x)


class RegressorAdapter(BaseModelAdapter):
    def __init__(self, model: object, fit_params: Dict[str, object] = None) -> None:
        super(RegressorAdapter, self).__init__(model, fit_params)

    def _underlying_predict(self, x: np.array) -> np.array:
        return self.model.predict(x)


class TSAdapter(BaseModelAdapter):
    def __init__(self, model: object, fit_params: Dict[str, object] = None) -> None:
        super(TSAdapter, self).__init__(model, fit_params)

    def _underlying_predict(self, x: np.array) -> np.array:
        return self.model.predict(x)


class CachedRegressorAdapter(RegressorAdapter):
    def __init__(self, model, fit_params=None):
        super(CachedRegressorAdapter, self).__init__(model, fit_params)
        self.prediction_cache = None

    def fit(self, x=None, y=None):
        """ At this point, the predictor has already been trained, but this
        has to be called to setup some things in the nonconformist backend """
        pass

    def predict(self, x=None):
        """ Same as in .fit()
        :return: np.array (n_test, n_classes) with class probability estimates """
        return self.prediction_cache


class CachedClassifierAdapter(ClassifierAdapter):
    def __init__(self, model, fit_params=None):
        super(CachedClassifierAdapter, self).__init__(model, fit_params)
        self.prediction_cache = None
        self.tempscale = True

    def fit(self, x=None, y=None):
        """ At this point, the predictor has already been trained, but this
        has to be called to setup some things in the nonconformist backend """
        pass

    def predict(self, x=None):
        """ Same as in .fit()
        :return: np.array (n_test, n_classes) with class probability estimates """
        if self.tempscale:
            return t_softmax(self.prediction_cache, t=0.5)
        else:
            return self.prediction_cache


class CachedTSAdapter(TSAdapter):
    def __init__(self, model, fit_params=None):
        super(CachedTSAdapter, self).__init__(model, fit_params)
        self.prediction_cache = None

    def fit(self, x=None, y=None):
        pass

    def predict(self, x=None):
        return self.prediction_cache



import math
import torch
import torch.nn as nn
from lightwood.helpers.torch import LightwoodAutocast


def len_to_mask(lengths, zeros):
    """
    :param lengths: list of ints with the lengths of the sequences
    :param zeros: bool. If false, the first lengths[i] values will be True and the rest will be false.
            If true, the first values will be False and the rest True
    :return: Boolean tensor of dimension (L, T) with L = len(lenghts) and T = lengths.max(), where with rows with lengths[i] True values followed by lengths.max()-lengths[i] False values. The True and False values are inverted if `zeros == True`
    """  # noqa
    # Clean trick from:
    # https://stackoverflow.com/questions/53403306/how-to-batch-convert-sentence-lengths-to-masks-in-pytorch
    mask = torch.arange(lengths.max(), device=lengths.device)[None, :] < lengths[:, None]
    if zeros:
        mask = ~mask  # Logical not
    return mask


def get_chunk(source, source_lengths, start, step):
    """Source is 3D tensor, shaped (batch_size, timesteps, n_dimensions), assuming static sequence length"""
    # Compute the lengths of the sequences (-1 due to the last element being used as target but not as data!
    trunc_seq_len = int(source_lengths[0].item() - 1)
    lengths = torch.zeros(source.shape[0]).fill_(trunc_seq_len).to(source.device)

    # This is necessary for MultiHeadedAttention to work
    end = min(start + step, trunc_seq_len)
    data = source[:, start:end, :]
    target = source[:, start + 1:end + 1, :]

    return data, target, lengths


class PositionalEncoding(nn.Module):
    def __init__(self, d_model, dropout=0.2, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        if not d_model % 2:
            pe[:, 1::2] = torch.cos(position * div_term)
        else:
            pe[:, 1::2] = torch.cos(position * div_term)[:, :d_model // 2]
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer("pe", pe)

    def forward(self, x):
        with LightwoodAutocast():
            x = x + self.pe[: x.size(0), :]
            return self.dropout(x)


class TransformerEncoder(nn.Module):
    def __init__(self, ninp, nhead, nhid, nlayers, dropout=0.2):
        super(TransformerEncoder, self).__init__()
        self.src_mask = None
        self.src_linear = nn.Linear(ninp, nhid)
        self.pos_encoder = PositionalEncoding(nhid, dropout)
        encoder_layers = nn.TransformerEncoderLayer(d_model=nhid, nhead=nhead, dim_feedforward=nhid, dropout=dropout)
        self.transformer_encoder = nn.TransformerEncoder(encoder_layers, nlayers)
        self.src_decoder = nn.Linear(nhid, ninp)
        self.init_weights()

    def _generate_square_subsequent_mask(self, sz):
        mask = (torch.triu(torch.ones(sz, sz)) == 1).transpose(0, 1)
        # Take the logarithm
        mask = (
            mask.float()
            .masked_fill(mask == 0, float("-inf"))
            .masked_fill(mask == 1, float(0.0))
        )
        return mask

    def init_weights(self):
        for p in self.parameters():
            if p.dim() > 1:
                torch.nn.init.xavier_uniform_(p)

    def forward(self, src, lengths, device):
        with LightwoodAutocast():
            if self.src_mask is None or self.src_mask.size(0) != src.size(0):
                # Attention mask to avoid attending to upcoming parts of the sequence
                self.src_mask = self._generate_square_subsequent_mask(src.size(0)).to(
                    device
                )
            src = self.src_linear(src)
            # src = self.pos_encoder(src) # not sure if this is helpful in time series
            # The lengths_mask has to be of size [batch, lengths]
            lengths_mask = len_to_mask(lengths, zeros=True).to(device)
            hidden = self.transformer_encoder(
                src, mask=self.src_mask, src_key_padding_mask=lengths_mask
            )
            output = self.src_decoder(hidden)
            return output, hidden

    def bptt(self, batch, criterion, device):
        """This method implements truncated backpropagation through time
        Returns: output tensor, None as hidden_state, which does not apply in this case, and loss value"""
        loss = 0
        train_batch, len_batch = batch
        batch_size, timesteps, _ = train_batch.shape

        for start_chunk in range(0, timesteps, timesteps):
            data, targets, lengths_chunk = get_chunk(train_batch, len_batch, start_chunk, timesteps)
            # Transformer expects seq_length in first dimension, so we transpose
            data = data.transpose(0, 1)
            targets = targets.transpose(0, 1)
            output, hidden = self.forward(data, lengths_chunk, device)
            loss += criterion(output, targets, lengths_chunk)

        return output.transpose(0, 1), hidden, loss



"""
2021.03.05

Basic helper functions for PretrainedLangEncoder
"""
import torch


class TextEmbed(torch.utils.data.Dataset):
    """
    Dataset class for quick embedding/label retrieval.
    Labels should be in the index space.

    If the labels provided are not in torch form, will convert them.
    """

    def __init__(self, encodings, labels):
        self.encodings = encodings
        self.labels = labels

    def __getitem__(self, idx):
        item = {key: torch.tensor(val[idx]) for key, val in self.encodings.items()}
        item['labels'] = self.labels[idx]
        return item

    def __len__(self):
        return len(self.labels)



"""
Time series utility methods for usage within mixers.
"""
from typing import Dict
from copy import deepcopy

import pandas as pd


def _transform_target(ts_analysis: Dict[str, Dict], df: pd.DataFrame, freq, group: tuple):
    df = deepcopy(df)  # needed because methods like get_level_values actually have undisclosed side effects  # noqa
    transformer = ts_analysis['stl_transforms'][group]['transformer']
    if isinstance(df.index, pd.MultiIndex) and len(df.index.levels) > 1:
        temp_s = df.droplevel(0)
        temp_s.index = pd.date_range(start=temp_s.index[0], freq=freq, periods=len(temp_s))
        return transformer.transform(temp_s.to_period())
    elif isinstance(df.index, pd.MultiIndex):
        temp_s = pd.Series(data=df.values)
        temp_s.index = df.index.levels[0]  # warning: do not use get_level_values as it removes inferred freq
        return transformer.transform(temp_s.to_period())
    else:
        return transformer.transform(df.to_period())


def _inverse_transform_target(ts_analysis: Dict[str, Dict], predictions: pd.DataFrame, freq, group: tuple):
    predictions = deepcopy(predictions)  # needed because methods like get_level_values actually have undisclosed side effects  # noqa
    transformer = ts_analysis['stl_transforms'][group]['transformer']
    if isinstance(predictions.index, pd.MultiIndex) and len(predictions.index.levels) > 1:
        temp_s = predictions.droplevel(0)
        temp_s.index = pd.date_range(start=temp_s.index[0], freq=freq, periods=len(temp_s))
        return transformer.inverse_transform(temp_s.to_period())
    elif isinstance(predictions.index, pd.MultiIndex):
        temp_s = pd.Series(data=predictions)
        temp_s.index = predictions.index.levels[0]  # warning: do not use get_level_values as it removes inferred freq
        return transformer.inverse_transform(temp_s.to_period())
    else:
        return transformer.inverse_transform(predictions.to_period())



from typing import List, Optional

import pandas as pd

from lightwood.mixer.base import BaseMixer
from lightwood.ensemble.base import BaseEnsemble
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs
from type_infer.dtype import dtype


class MeanEnsemble(BaseEnsemble):
    """
    When called, this ensemble will return the mean prediction from the entire list of underlying mixers.

    NOTE: can only be used in regression tasks.
    """
    def __init__(self, target, mixers: List[BaseMixer], data: EncodedDs, dtype_dict: dict,
                 fit: Optional[bool] = True, **kwargs) -> None:
        super().__init__(target, mixers, data, fit=False)
        if dtype_dict[target] not in (dtype.float, dtype.integer, dtype.quantity):
            raise Exception(
                f'This ensemble can only be used regression problems! Got target dtype {dtype_dict[target]} instead!')

    def __call__(self, ds: EncodedDs, args: PredictionArguments) -> pd.DataFrame:
        predictions_df = pd.DataFrame()
        for mixer in self.mixers:
            predictions_df[f'__mdb_mixer_{type(mixer).__name__}'] = mixer(ds, args=args)['prediction']

        return pd.DataFrame(predictions_df.mean(axis='columns'), columns=['prediction'])




from typing import Optional
import pandas as pd

from lightwood.data.encoded_ds import EncodedDs
from lightwood.api.types import PredictionArguments


class BaseMixer:
    """
    Base class for all mixers.

    Mixers are the backbone of all Lightwood machine learning models. They intake encoded feature representations for every column, and are tasked with learning to fulfill the predictive requirements stated in a problem definition.
    
    There are two important methods for any mixer to work:
        1. `fit()` contains all logic to train the mixer with the training data that has been encoded by all the (already trained) Lightwood encoders for any given task.
        2. `__call__()` is executed to generate predictions once the mixer has been trained using `fit()`. 
    
    An additional `partial_fit()` method is used to update any mixer that has already been trained.

    Class Attributes:
    - stable: If set to `True`, this mixer should always work. Any mixer with `stable=False` can be expected to fail under some circumstances.
    - fit_data_len: Length of the training data.
    - supports_proba: For classification tasks, whether the mixer supports yielding per-class scores rather than only returning the predicted label. 
    - trains_once: If True, the mixer is trained once during learn, using all available input data (`train` and `dev` splits for training, `test` for validation). Otherwise, it trains once with the `train`` split & `dev` for validation, and optionally (depending on the problem definition `fit_on_all` and mixer-wise `fit_on_dev` arguments) a second time after post-training analysis via partial_fit, with `train` and `dev` splits as training subset, and `test` split as validation. Should only be set to True for mixers that don't require post-training analysis, as otherwise actual validation data would be treated as a held-out portion, which is a mistake. 
    """  # noqa
    stable: bool
    fit_data_len: int  # @TODO (Patricio): should this really be in `BaseMixer`?
    supports_proba: bool
    trains_once: bool

    def __init__(self, stop_after: float):
        """
        :param stop_after: Time budget (in seconds) to train this mixer.
        """
        self.stop_after = stop_after
        self.supports_proba = False
        self.trains_once = False

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        Fits/trains a mixer with training data. 
         
        :param train_data: encoded representations of the training data subset. 
        :param dev_data: encoded representations of the "dev" data subset. This can be used as an internal validation subset (e.g. it is used for early stopping in the default `Neural` mixer). 
         
        """  # noqa
        raise NotImplementedError()

    def __call__(self, ds: EncodedDs,
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        """
        Calls a trained mixer to predict the target column given some input data.
        
        :param ds: encoded representations of input data.
        :param args: a `lightwood.api.types.PredictionArguments` object, including all relevant inference-time arguments to customize the behavior.
        :return: 
        """  # noqa
        raise NotImplementedError()

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs, adjust_args: Optional[dict] = None) -> None:
        """
        Partially fits/trains a mixer with new training data. This is a somewhat experimental method, and it aims at updating pre-existing Lightwood predictors. 

        :param train_data: encoded representations of the new training data subset. 
        :param dev_data: encoded representations of new the "dev" data subset. As in `fit()`, this can be used as an internal validation subset. 
        :param adjust_args: optional arguments to customize the finetuning process.

        """  # noqa
        pass



from lightwood.api.types import (
    JsonAI,
    ProblemDefinition,
    TimeseriesSettings,
    ModelAnalysis,
    PredictionArguments,
)
from lightwood.api.predictor import PredictorInterface
from lightwood.api.high_level import (
    code_from_problem,
    predictor_from_problem,
    predictor_from_code,
    code_from_json_ai,
    json_ai_from_problem,
    predictor_from_state,
    load_custom_module,
)

__all__ = [
    "code_from_problem",
    "predictor_from_problem",
    "predictor_from_code",
    "code_from_json_ai",
    "json_ai_from_problem",
    "JsonAI",
    "ProblemDefinition",
    "TimeseriesSettings",
    "ModelAnalysis",
    "PredictionArguments",
    "PredictorInterface",
    "predictor_from_state",
    "load_custom_module",
]



from typing import Dict, List, Union, Optional
from copy import deepcopy

import numpy as np
import pandas as pd

from lightwood.helpers.log import log
from lightwood.encoder.base import BaseEncoder
from lightwood.mixer.base import BaseMixer
from lightwood.mixer.xgboost import XGBoostMixer
from lightwood.api.types import PredictionArguments, TimeseriesSettings
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs


class XGBoostArrayMixer(BaseMixer):
    """XGBoost-based model, intended for usage in forecasting tasks."""
    models: List[XGBoostMixer]
    submodel_stop_after: float
    target: str
    supports_proba: bool
    ts_analysis: Dict
    tss: TimeseriesSettings

    def __init__(
            self,
            stop_after: float,
            target: str,
            dtype_dict: Dict[str, str],
            input_cols: List[str],
            fit_on_dev: bool,
            target_encoder: BaseEncoder,
            ts_analysis: Dict[str, object],
            use_stl: bool,
            tss: TimeseriesSettings
    ):
        super().__init__(stop_after)
        self.tss = tss
        self.horizon = tss.horizon
        self.submodel_stop_after = stop_after / self.horizon
        self.target = target
        self.offset_pred_cols = [f'{self.target}_timestep_{i}' for i in range(1, self.horizon)]
        if set(input_cols) != {self.tss.order_by}:
            input_cols.remove(self.tss.order_by)
        for col in self.offset_pred_cols:
            dtype_dict[col] = dtype_dict[self.target]
        self.models = [XGBoostMixer(self.submodel_stop_after,
                                    target_col,
                                    dtype_dict,
                                    input_cols,
                                    False,  # fit_on_dev,
                                    False,  # use_optuna
                                    target_encoder)
                       for _, target_col in zip(range(self.horizon), [target] + self.offset_pred_cols)]
        self.ts_analysis = ts_analysis
        self.supports_proba = False
        self.use_stl = False
        self.stable = False

    def _fit(self, train_data: EncodedDs, dev_data: EncodedDs, submodel_method='fit') -> None:
        original_train = deepcopy(train_data.data_frame)
        original_dev = deepcopy(dev_data.data_frame)

        for timestep in range(self.horizon):
            getattr(self.models[timestep], submodel_method)(train_data, dev_data)

        # restore dfs
        train_data.data_frame = original_train
        dev_data.data_frame = original_dev

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        log.info('Started fitting LGBM models for array prediction')
        self._fit(train_data, dev_data, submodel_method='fit')

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs, args: Optional[dict] = None) -> None:
        log.info('Updating array of LGBM models...')
        self._fit(train_data, dev_data, submodel_method='partial_fit')

    def __call__(self, ds: Union[EncodedDs, ConcatedEncodedDs],
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        if args.predict_proba:
            log.warning('This model does not output probability estimates')

        original_df = deepcopy(ds.data_frame)
        length = sum(ds.encoded_ds_lengths) if isinstance(ds, ConcatedEncodedDs) else len(ds)
        ydf = pd.DataFrame(0,  # zero-filled
                           index=np.arange(length),
                           columns=[f'prediction_{i}' for i in range(self.horizon)])

        for timestep in range(self.horizon):
            ydf[f'prediction_{timestep}'] = self.models[timestep](ds, args)['prediction'].values

        if self.models[0].positive_domain:
            ydf = ydf.clip(0)

        ydf['prediction'] = ydf.values.tolist()
        ds.data_frame = original_df
        return ydf[['prediction']]



import os
import logging
import colorlog
from time import time
from datetime import datetime
from functools import wraps


def initialize_log():
    pid = os.getpid()

    handler = colorlog.StreamHandler()
    handler.setFormatter(colorlog.ColoredFormatter())

    logging.basicConfig(handlers=[handler])
    log = logging.getLogger(f'lightwood-{pid}')
    log_level = os.environ.get('LIGHTWOOD_LOG', 'DEBUG')
    log.setLevel(log_level)
    return log


def timed_predictor(f):
    """
    Intended to be called from within lightwood predictor methods.
    We use `wraps` to pass metadata into debuggers (as in stackoverflow.com/a/27737385)
    """
    @wraps(f)
    def wrap(predictor, *args, **kw):
        ts = time()
        result = f(predictor, *args, **kw)
        te = time()
        log.debug(f' `{f.__name__}` runtime: {round(te - ts, 2)} seconds')
        if hasattr(predictor, 'runtime_log'):
            predictor.runtime_log[(f.__name__, datetime.fromtimestamp(ts))] = round(te - ts, 2)
        return result
    return wrap


def timed(f):
    """
    Intended to be called from within any lightwood method to log the runtime.
    """
    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        log.debug(f' `{f.__name__}` runtime: {round(te - ts, 2)} seconds')
        return result
    return wrap


log = initialize_log()



from copy import deepcopy
from typing import Dict, List, Union, Optional

import numpy as np
import pandas as pd

from lightwood.helpers.log import log
from lightwood.encoder.base import BaseEncoder
from lightwood.mixer.base import BaseMixer
from lightwood.mixer.lightgbm import LightGBM
from lightwood.api.types import PredictionArguments, TimeseriesSettings
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs


class LightGBMArray(BaseMixer):
    """LightGBM-based model, intended for usage in time series tasks."""
    models: List[LightGBM]
    submodel_stop_after: float
    target: str
    supports_proba: bool
    ts_analysis: Dict
    tss: TimeseriesSettings

    def __init__(
            self,
            stop_after: float,
            target: str,
            dtype_dict: Dict[str, str],
            input_cols: List[str],
            fit_on_dev: bool,
            target_encoder: BaseEncoder,
            ts_analysis: Dict[str, object],
            use_stl: bool,
            tss: TimeseriesSettings
    ):
        super().__init__(stop_after)
        self.tss = tss
        self.horizon = tss.horizon
        self.submodel_stop_after = stop_after / self.horizon
        self.target = target
        self.offset_pred_cols = [f'{self.target}_timestep_{i}' for i in range(1, self.horizon)]
        if set(input_cols) != {self.tss.order_by}:
            input_cols.remove(self.tss.order_by)
        for col in self.offset_pred_cols:
            dtype_dict[col] = dtype_dict[self.target]
        self.models = [LightGBM(self.submodel_stop_after,
                                target_col,
                                dtype_dict,
                                input_cols,
                                False,  # fit_on_dev,
                                True if tss.horizon < 10 else False,  # use_optuna
                                target_encoder)
                       for _, target_col in zip(range(self.horizon), [target] + self.offset_pred_cols)]
        self.ts_analysis = ts_analysis
        self.supports_proba = False
        self.use_stl = False
        self.stable = True

    def _fit(self, train_data: EncodedDs, dev_data: EncodedDs, submodel_method='fit') -> None:
        original_train = deepcopy(train_data.data_frame)
        original_dev = deepcopy(dev_data.data_frame)

        for timestep in range(self.horizon):
            getattr(self.models[timestep], submodel_method)(train_data, dev_data)

        # restore dfs
        train_data.data_frame = original_train
        dev_data.data_frame = original_dev

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        log.info('Started fitting LGBM models for array prediction')
        self._fit(train_data, dev_data, submodel_method='fit')

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs, args: Optional[dict] = None) -> None:
        log.info('Updating array of LGBM models...')
        self._fit(train_data, dev_data, submodel_method='partial_fit')

    def __call__(self, ds: Union[EncodedDs, ConcatedEncodedDs],
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        if args.predict_proba:
            log.warning('This model does not output probability estimates')

        original_df = deepcopy(ds.data_frame)
        length = sum(ds.encoded_ds_lengths) if isinstance(ds, ConcatedEncodedDs) else len(ds)
        ydf = pd.DataFrame(0,  # zero-filled
                           index=np.arange(length),
                           columns=[f'prediction_{i}' for i in range(self.horizon)])

        for timestep in range(self.horizon):
            ydf[f'prediction_{timestep}'] = self.models[timestep](ds, args)['prediction'].values

        if self.models[0].positive_domain:
            ydf = ydf.clip(0)

        ydf['prediction'] = ydf.values.tolist()
        ds.data_frame = original_df
        return ydf[['prediction']]



import time
from math import gcd
from typing import List
from copy import deepcopy

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from torch import optim

from type_infer.dtype import dtype
from lightwood.helpers.log import log
from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.device import get_device_from_name
from lightwood.helpers.torch import LightwoodAutocast
from lightwood.encoder.datetime import DatetimeNormalizerEncoder
from lightwood.encoder.time_series.helpers.rnn_helpers import EncoderRNNNumerical, DecoderRNNNumerical
from lightwood.encoder.helpers import MinMaxNormalizer, CatNormalizer
from lightwood.encoder.time_series.helpers.transformer_helpers import TransformerEncoder, get_chunk, len_to_mask


class TimeSeriesEncoder(BaseEncoder):
    """
    Time series encoder. This module can learn features for any `order_by` temporal column, both with and without accompanying target data.

    The backbone of this encoder is either a recurrent neural network or a transformer; both structured in an encoder-decoder fashion.
    """  # noqa
    is_timeseries_encoder: bool = True
    is_trainable_encoder: bool = True

    def __init__(self,
                 stop_after: float,
                 is_target: bool = False,
                 original_type: str = None,
                 target: str = None,
                 grouped_by: List[str] = [],
                 encoder_type: str = 'rnn',
                 device: str = ''):
        super().__init__(is_target)
        self.device = get_device_from_name(device)
        self.target = target
        self.grouped_by = grouped_by
        self._learning_rate = 0.01
        self.output_size = 128
        self._transformer_hidden_size = None
        self._epochs = int(1e5)  # default training epochs
        self._stop_on_n_bad_epochs = 5  # stop training after N epochs where loss is worse than running avg
        self._epochs_running_avg = 5  # amount of epochs for running average
        self._pytorch_wrapper = torch.FloatTensor
        self.is_prepared = False
        self._is_setup = False
        self._max_ts_length = 0
        self._sos = 0.0  # start of sequence for decoding
        self._eos = 0.0  # end of input sequence -- padding value for batches
        self._n_dims = 1
        self._normalizer = None
        self.dep_norms = {}  # dict of dict of normalizers for each dependency (can be grouped-by some column)
        self._target_type = None
        self._group_combinations = None
        self.original_type = original_type
        self.stop_after = stop_after
        if encoder_type.lower() == 'rnn':
            self.encoder_class = EncoderRNNNumerical
        elif encoder_type.lower() == 'transformer':
            self.encoder_class = TransformerEncoder

    def setup_nn(self, ts_analysis, dependencies=None):
        """This method must be executed after initializing, else types are unassigned"""
        if self.original_type in (dtype.datetime, dtype.date):
            self._normalizer = DatetimeNormalizerEncoder(sinusoidal=True)
            self._n_dims *= len(self._normalizer.fields) * 2  # sinusoidal datetime components
        elif self.original_type in (dtype.float, dtype.integer):
            self._normalizer = MinMaxNormalizer()

        total_dims = self._n_dims
        dec_hsize = self.output_size

        if dependencies:
            for dep_name, dep in dependencies.items():
                self.dependencies.append(dep_name)

                if dep_name in self.grouped_by:
                    continue  # we only use group column for indexing and selecting rows

                assert dep['original_type'] in (dtype.categorical, dtype.binary, dtype.cat_tsarray,
                                                dtype.integer, dtype.float, dtype.num_tsarray)

                if f'__mdb_ts_previous_{self.target}' == dep_name:
                    self.dep_norms[dep_name] = ts_analysis['target_normalizers']
                    self._group_combinations = ts_analysis['group_combinations']
                    self._target_type = dep['original_type']

                # if TS analysis yields no normalizers for this dependency, we create a generic one based on its dtype
                else:
                    if dep['original_type'] in (dtype.categorical, dtype.binary):
                        self.dep_norms[dep_name]['__default'] = CatNormalizer()
                    else:
                        self.dep_norms[dep_name]['__default'] = MinMaxNormalizer()

                    self.dep_norms[dep_name]['__default'].prepare(dep['data'])
                    self._group_combinations = {'__default': None}

                # add descriptor size to the total encoder output dimensionality
                if dep['original_type'] in (dtype.categorical, dtype.binary):
                    total_dims += len(self.dep_norms[dep_name]['__default'].scaler.categories_[0])
                elif dep['original_type'] in (dtype.integer, dtype.float, dtype.num_tsarray, dtype.cat_tsarray):
                    total_dims += 1

        if self.encoder_class == EncoderRNNNumerical:
            self._enc_criterion = nn.MSELoss()
            self._dec_criterion = self._enc_criterion
            self._encoder = self.encoder_class(input_size=total_dims,
                                               hidden_size=self.output_size).to(self.device)
        elif self.encoder_class == TransformerEncoder:
            self._enc_criterion = self._masked_criterion
            self._dec_criterion = nn.MSELoss()
            self._base_criterion = nn.MSELoss(reduction="none")
            if self._transformer_hidden_size is None:
                self._transformer_hidden_size = total_dims * 2  # arbitrary

            self._encoder = self.encoder_class(ninp=total_dims,
                                               nhead=gcd(dec_hsize, total_dims),
                                               nhid=self._transformer_hidden_size,
                                               nlayers=1).to(self.device)
        else:
            raise Exception(f"Time series encoder class not supported: {self.encoder_class}")

        self._decoder = DecoderRNNNumerical(output_size=total_dims, hidden_size=dec_hsize).to(self.device)
        self._parameters = list(self._encoder.parameters()) + list(self._decoder.parameters())
        self._optimizer = optim.AdamW(self._parameters, lr=self._learning_rate, weight_decay=1e-4)
        self._n_dims = total_dims
        self._is_setup = True

    def to(self, device, available_devices):
        if self._is_setup:
            self.device = device
            return super().to(device, available_devices)
        return self

    def _prepare_raw_data(self, data):
        """Convert to array and determine max length"""
        out_data = []
        for e in data:
            if not isinstance(e, torch.Tensor):
                e = np.array(e, dtype=float)
                e[np.isnan(e)] = 0.0
                t = torch.tensor(e, dtype=torch.float)
            else:
                t = e.float()
            t[torch.isnan(t)] = 0.0
            out_data.append(t)
        lengths = torch.tensor([len(e) for e in data], dtype=torch.float)
        return out_data, lengths

    def _get_batch(self, source, start, end):
        end = min(end, len(source))
        return source[start:end]

    def prepare(self, train_priming_data: pd.Series, dev_priming_data: pd.Series, dependency_data={}, ts_analysis=None,
                feedback_hoop_function=log.info, batch_size=256):
        """
        :param priming_data: a list of (self._n_dims)-dimensional time series [[dim1_data], ...]
        :param dependency_data: raw data from other columns
        :param ts_analysis: dictionary with time analysis info (e.g. normalizers for each target group)
        :param feedback_hoop_function: method to use if you want to get feedback on the training process
        :param batch_size
        """
        priming_data = pd.concat([train_priming_data, dev_priming_data])
        priming_data = list(priming_data.values)

        if self.is_prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')
        else:
            self.setup_nn(ts_analysis, dependency_data)

        started = time.time()

        # Convert to array and determine max length
        priming_data, lengths_data = self._prepare_raw_data(priming_data)
        self._max_ts_length = int(lengths_data.max())

        if self._normalizer:
            self._normalizer.prepare(priming_data)
            priming_data = self._normalizer.encode(priming_data).to(self.device)
            if len(priming_data.shape) < 3:
                priming_data = priming_data.unsqueeze(-1)
        else:
            priming_data = torch.stack([d for d in priming_data]).unsqueeze(-1).to(self.device)

        # merge all normalized data into a training batch
        normalized_tensors = []
        for dep_name, dep_data in dependency_data.items():
            if dep_name in self.grouped_by:
                continue
            if dep_data['original_type'] in (dtype.integer, dtype.float):
                dep_data['group_info'] = {group: dependency_data[group]['data'] for group in self.grouped_by}
                data = torch.zeros((len(priming_data), lengths_data.max().int().item(), 1))
                all_idxs = set(range(len(data)))
                grouped = dep_data['data'].groupby(by=self.grouped_by)
                for group_name, normalizer in self.dep_norms[dep_name].items():
                    if group_name != '__default':
                        subset = grouped.get_group(normalizer.combination)
                        normalized = normalizer.encode(subset).unsqueeze(-1)
                        data[subset.index, :, :] = normalized
                        all_idxs -= set(subset.index)
                if len(all_idxs) > 0 and '__default' in self.dep_norms[dep_name].keys():
                    default_norm = self.dep_norms[dep_name]['__default']
                    subset = [dep_data['data'][idx] for idx in list(all_idxs)]
                    data[list(all_idxs), :, :] = torch.Tensor(default_norm.encode(subset)).unsqueeze(-1)

            else:
                # categorical has only one normalizer at all times
                normalizer = self.dep_norms[dep_name]['__default']
                data = normalizer.encode(dep_data['data'].values)
                if len(data.shape) < 3:
                    data = data.unsqueeze(-1)  # add feature dimension
            data[torch.isnan(data)] = 0.0
            normalized_tensors.append(data)

        if normalized_tensors:
            normalized_data = torch.cat(normalized_tensors, dim=-1).to(self.device)
            priming_data = torch.cat([priming_data, normalized_data], dim=-1)

        self._encoder.train()
        running_losses = np.full(self._epochs_running_avg, np.nan)
        bad_epochs = 0

        for epoch in range(self._epochs):
            average_loss = 0

            for batch_idx in range(0, len(priming_data), batch_size):
                # setup loss and optimizer
                self._optimizer.zero_grad()
                loss = 0

                # shape: (batch_size, timesteps, n_dims)
                batch = self._get_batch(priming_data, batch_idx, min(batch_idx + batch_size, len(priming_data)))

                # encode and decode through time
                with LightwoodAutocast():
                    if self.encoder_class == TransformerEncoder:
                        # pack batch length info tensor
                        len_batch = self._get_batch(lengths_data, batch_idx, min(
                            batch_idx + batch_size, len(priming_data)))
                        batch = batch, len_batch

                        next_tensor, hidden_state, dec_loss = self._encoder.bptt(
                            batch, self._enc_criterion, self.device)
                        loss += dec_loss

                    else:
                        next_tensor, hidden_state, enc_loss = self._encoder.bptt(
                            batch, self._enc_criterion, self.device)
                        loss += enc_loss

                        next_tensor, hidden_state, dec_loss = self._decoder.decode(
                            batch, next_tensor, self._dec_criterion, self.device, hidden_state=hidden_state)
                        loss += dec_loss

                loss.backward()

                self._optimizer.step()
                average_loss += loss.item()

            average_loss = average_loss / len(priming_data)
            batch_idx += batch_size

            if epoch > self._epochs_running_avg and average_loss > np.average(running_losses):
                bad_epochs += 1

            # update running loss
            running_losses[:-1] = running_losses[1:]
            running_losses[-1] = average_loss

            if feedback_hoop_function is not None:
                feedback_hoop_function(
                    "time series encoder epoch [{epoch_n}/{total}] average_loss = {average_loss}".format(
                        epoch_n=epoch + 1, total=self._epochs, average_loss=average_loss))

            if bad_epochs > self._stop_on_n_bad_epochs:
                break
            elif (time.time() - started) > self.stop_after:
                break

        self.is_prepared = True

    def _encode_one(self, data, previous=None, initial_hidden=None, return_next_value=False):
        """
        This method encodes one single row of serial data
        :param data: multidimensional time series as list of lists [[dim1_data], [dim2_data], ...]
                     (dim_data: string with format "x11, x12, ... x1n")
        :param initial_hidden: if you want to encode from an initial hidden state other than 0s
        :param return_next_value:  if you want to return the next value in the time series too

        :return:  either encoded_value or (encoded_value, next_value)
        """
        self._encoder.eval()
        with torch.no_grad():
            # Convert to array and determine max length
            data, lengths_data = self._prepare_raw_data(data)
            self._max_ts_length = int(lengths_data.max())

            if self._normalizer:
                data = self._normalizer.encode(data).to(self.device)
                if len(data.shape) < 3:
                    data = data.unsqueeze(-1)
            else:
                data = torch.stack([d for d in data]).unsqueeze(-1).to(self.device)

            if previous is not None:
                target_tensor = torch.stack(previous).to(self.device)
                target_tensor[torch.isnan(target_tensor)] = 0.0
                if len(target_tensor.shape) < 3:
                    target_tensor = target_tensor.transpose(0, 1).unsqueeze(0)
                data_tensor = torch.cat((data, target_tensor), dim=-1)
            else:
                data_tensor = data

            steps = data_tensor.shape[1]

            if self.encoder_class == EncoderRNNNumerical:
                encoder_hidden = self._encoder.init_hidden(self.device)
                encoder_hidden = encoder_hidden if initial_hidden is None else initial_hidden

                next_tensor = None
                for tensor_i in range(steps):
                    next_tensor, encoder_hidden = self._encoder.forward(data_tensor[:, tensor_i, :].unsqueeze(dim=0),
                                                                        encoder_hidden)

            else:
                next_tensor = None
                len_batch = self._get_batch(lengths_data, 0, len(data))
                batch_size, timesteps, _ = data_tensor.shape

                for start_chunk in range(0, timesteps, timesteps):
                    data, targets, lengths_chunk = get_chunk(data_tensor, len_batch, start_chunk, timesteps)
                    data = data.transpose(0, 1)
                    next_tensor, encoder_hidden = self._encoder.forward(data, lengths_chunk, self.device)

        if return_next_value:
            return encoder_hidden, next_tensor
        else:
            return encoder_hidden

    def encode(self, column_data, dependency_data=None, get_next_count=None):
        """
        Encode a list of time series data
        :param column_data: a list of (self._n_dims)-dimensional time series [[dim1_data], ...] to encode
        :param get_next_count: default None, but you can pass a number X and it will return the X following predictions
                               on the series for each ts_data_point in column_data
        :return: a list of encoded time series or if get_next_count !=0 two lists (encoded_values, projected_numbers)
        """

        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        if isinstance(column_data, pd.Series):
            data = deepcopy(column_data.values)  # get a copy to avoid modifying the actual data frame
        else:
            data = column_data

        for i in range(len(data)):
            if not isinstance(data[i][0], list):
                data[i] = [data[i]]  # add dimension for 1D timeseries

        # include autoregressive target data
        ptd = []
        if dependency_data is not None:
            for dep, dep_data in dependency_data.items():
                if dep in self.grouped_by:
                    continue
                # normalize numerical target per group-by
                if self._target_type in (dtype.integer, dtype.float, dtype.num_tsarray):
                    dep_info = {
                        'group_info': {group: dependency_data[group] for group in self.grouped_by},
                        'data': dep_data
                    }
                    tensor = torch.zeros((len(dep_data), len(dep_data[0]), 1)).to(self.device)
                    all_idxs = set(range(len(dep_data)))

                    grouped = dep_info['data'].groupby(by=self.grouped_by)
                    for combination in [c for c in self._group_combinations if c != '__default']:
                        normalizer = self.dep_norms[dep].get(tuple(combination), None)
                        if normalizer is None:
                            normalizer = self.dep_norms[dep]['__default']
                        subdf = grouped.get_group(normalizer.combination)
                        tensor[subdf.index, :, :] = torch.Tensor(normalizer.encode(subdf)).unsqueeze(-1).to(self.device)
                        all_idxs -= set(subdf.index)

                    # encode all remaining rows (not belonging to any grouped combination) with default normalizer
                    if all_idxs:
                        default_norm = self.dep_norms[dep]['__default']
                        subset = [dep_data[idx] for idx in all_idxs]
                        tensor[list(all_idxs), :, :] = torch.Tensor(
                            default_norm.encode(subset)).unsqueeze(-1).to(self.device)
                        tensor[torch.isnan(tensor)] = 0.0

                # normalize categorical target
                else:
                    normalizer = self.dep_norms[dep]['__default']
                    tensor = normalizer.encode(dep_data)
                    tensor[torch.isnan(tensor)] = 0.0

                ptd.append(tensor)

        ret = []
        next = []

        for i, val in enumerate(data):
            if get_next_count is None:
                if dependency_data is not None and len(dependency_data) > 0 and len(ptd) > 0:
                    encoded = self._encode_one(val, previous=[values[i] for values in ptd])
                else:
                    encoded = self._encode_one(val)

            else:
                if get_next_count <= 0:
                    raise Exception('get_next_count must be greater than 0')

                hidden = None
                vector = val
                next_i = []

                for j in range(get_next_count):
                    hidden, next_reading = self._encode_one(vector, initial_hidden=hidden, return_next_value=True)
                    vector = [next_reading]
                    if j == 0:
                        encoded = hidden
                    next_i.append(next_reading)

                next_value = next_i[0][0].cpu()

                if self._normalizer:
                    next_value = torch.Tensor(self._normalizer.decode(next_value))

                next.append(next_value)

            ret.append(encoded[0][0].cpu())

        if get_next_count is None:
            return torch.stack(ret)
        else:
            return torch.stack(ret), torch.stack(next)

    def _decode_one(self, hidden, steps):
        """
        Decodes a single time series from its encoded representation.
        :param hidden: time series embedded representation tensor, with size self.output_size
        :param steps: as in decode(), defines how many values to output when reconstructing
        :return: decoded time series list
        """
        self._decoder.eval()
        with torch.no_grad():
            ret = []
            next_tensor = torch.full((1, 1, self._n_dims), self._sos, dtype=torch.float32).to(self.device)
            timesteps = steps if steps else self._max_ts_length
            for _ in range(timesteps):
                next_tensor, hidden = self._decoder.forward(next_tensor, hidden)
                ret.append(next_tensor)
            return torch.stack(ret)

    def decode(self, encoded_data, steps=None):
        """
        Decode a list of embedded multidimensional time series
        :param encoded_data: a list of embeddings [ e1, e2, ...] to be decoded into time series
        :param steps: fixed number of timesteps to reconstruct from each embedding.
        If None, encoder will output the largest length encountered during training.
        :return: a list of reconstructed time series
        """
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        ret = []
        for _, val in enumerate(encoded_data):
            hidden = torch.unsqueeze(torch.unsqueeze(val, dim=0), dim=0).to(self.device)
            reconstruction = self._decode_one(hidden, steps).cpu().squeeze().T.numpy()

            if self._n_dims == 1:
                reconstruction = reconstruction.reshape(1, -1)

            if self._normalizer:
                reconstruction = self._normalizer.decode(reconstruction)

            ret.append(reconstruction)

        return torch.Tensor(ret)

    def _masked_criterion(self, output, targets, lengths):
        """ Computes the loss of the first `lengths` items in the chunk """
        # Put in (B, T) format and zero-out the unnecessary values
        mask = len_to_mask(lengths, zeros=False).t()

        # Inflate to feature dimension
        mask = mask.unsqueeze(-1).repeat(1, 1, output.shape[-1])
        output = output * mask
        targets = targets * mask

        # compute the loss with respect to the appropriate lengths and average across the batch-size
        # We compute for every output (x_i)_i=1^L and target (y_i)_i=1^L, loss = 1/L \sum (x_i - y_i)^2
        # And average across the mini-batch
        losses = self._base_criterion(output, targets).sum(dim=2).sum(dim=0)

        # The TBPTT will compute a slightly different loss, but it is not problematic
        loss = torch.dot((1.0 / lengths.float()), losses) / len(losses)

        return loss



from typing import List, Iterable
import torch


class BaseEncoder:
    """
    Base class for all encoders.
    
    An encoder should return encoded representations of any columnar data.
    The procedure for this is defined inside the `encode()` method.
    
    If this encoder is expected to handle an output column, then it also needs to implement the respective `decode()` method that handles the inverse transformation from encoded representations to the final prediction in the original column space.
    
    For encoders that learn representations (as opposed to rule-based), the `prepare()` method will handle all learning logic.
    
    The `to()` method is used to move PyTorch-based encoders to and from a GPU.
    
    :param  is_target: Whether the data to encode is the target, as per the problem definition.
    :param is_timeseries_encoder: Whether encoder represents sequential/time-series data. Lightwood must provide specific treatment for this kind of encoder
    :param is_trainable_encoder: Whether the encoder must return learned representations. Lightwood checks whether this flag is present in order to pass data to the feature representation via the ``prepare`` statement. 
    
    Class Attributes:
    - is_prepared: Internal flag to signal that the `prepare()` method has been successfully executed.
    - is_nn_encoder: Whether the encoder is neural network-based.
    - dependencies: list of additional columns that the encoder might need to encode.
    - output_size: length of each encoding tensor for a single data point.
    
    """ # noqa
    is_target: bool
    is_prepared: bool

    is_timeseries_encoder: bool = False
    is_trainable_encoder: bool = False

    def __init__(self, is_target=False) -> None:
        self.is_target = is_target
        self.is_prepared = False
        self.dependencies = []
        self.output_size = None

    # Not all encoders need to be prepared
    def prepare(self, priming_data: Iterable[object]) -> None:
        """
        Given 'priming_data' (i.e. training data), prepares encoders either through a rule-based (ex: one-hot encoding) or learned (ex: DistilBERT for text) model. This works explicitly on only training data.

        :param priming_data: An iterable data structure where all the elements have type that is compatible with the encoder processing type; this may differ per encoder.
        """ # noqa
        self.is_prepared = True

    def encode(self, column_data: Iterable[object]) -> torch.Tensor:
        """
        Given the approach defined in `prepare()`, encodes column data into a numerical representation to form part of the feature vector.

        After all columns are featurized, each encoded vector is concatenated to form a feature vector per row in the dataset.

        :param column_data: An iterable data structure where all the elements have type that is compatible with the encoder processing type; this may differ per encoder.

        :returns: The encoded representation of data, per column
        """ # noqa
        raise NotImplementedError

    def decode(self, encoded_data: torch.Tensor) -> List[object]:
        """
        Given an encoded representation, returns the decoded value. Decoded values may not exist for all encoders (ex: rich text, audio, etc.)

        :param encoded_data: The input representation in encoded format

        :returns: The decoded representation of data, per column, in the original data-type presented.
        """ # noqa
        raise NotImplementedError

    # TODO Should work for all torch-based encoders, but custom behavior may have to be implemented for weird models
    def to(self, device, available_devices):
        # Find all nn.Module type objects and convert them
        # @TODO: Make this work recursively
        for v in vars(self):
            attr = getattr(self, v)
            if isinstance(attr, torch.nn.Module):
                attr.to(device)
        return self



from typing import List
import torch
from lightwood.encoder import BaseEncoder
from lightwood.encoder.categorical import CategoricalAutoEncoder
from type_infer.rule_based.helpers import tokenize_text
from lightwood.helpers.torch import concat_vectors_and_pad, average_vectors
import pandas as pd


class ShortTextEncoder(BaseEncoder):
    is_trainable_encoder = False

    def __init__(self, is_target=False, mode=None, device=''):
        """
        :param is_target:
        :param mode:
            None or "concat" or "mean".
            When None, it will be set automatically based on is_target:
            (is_target) -> 'concat'
            (not is_target) -> 'mean'
        :param device: name of the device that get_device_from_name will attempt to use.
        """
        super().__init__(is_target)

        if mode is None:
            if is_target:
                self._mode = 'concat'
            else:
                self._mode = 'mean'
        else:
            if mode not in ['concat', 'mean']:
                self._unexpected_mode()

            if is_target and mode != 'concat':
                raise ValueError('mode must be "concat" when is_target=True')

            self._mode = mode

        # Defined in self.prepare()
        self._combine_fn = None
        self.max_words_per_sent = None
        self.cae = CategoricalAutoEncoder(is_target=is_target, max_encoded_length=100, device=device)
        self.is_prepared = False

    def _unexpected_mode(self):
        raise ValueError('unexpected combine value (must be "mean" or "concat")')

    # defining both of these as normal functions because pickle can't deal with lambdas
    def _combine_concat(self, vecs):
        return concat_vectors_and_pad(vecs, self.max_words_per_sent)

    def _combine_mean(self, vecs):
        return average_vectors(vecs)

    def prepare(self, priming_data):
        no_null_sentences = (x if x is not None else '' for x in priming_data)
        unique_tokens = set()
        max_words_per_sent = 0
        for sent in no_null_sentences:
            tokens = list(tokenize_text(sent))
            max_words_per_sent = max(max_words_per_sent, len(tokens))
            for tok in tokens:
                unique_tokens.add(tok)

        self.cae.prepare(pd.Series(list(unique_tokens)), pd.Series([]))

        if self._mode == 'concat':
            self.max_words_per_sent = max_words_per_sent
            self._combine_fn = self._combine_concat
        elif self._mode == 'mean':
            self._combine_fn = self._combine_mean
        else:
            self._unexpected_mode()

        self.is_prepared = True
        encoded = self.encode([priming_data[0]])
        self.output_size = len(encoded[0])

    def encode(self, column_data: List[str]) -> torch.Tensor:
        no_null_sentences = (x if x is not None else '' for x in column_data)
        output = []
        for sent in no_null_sentences:
            tokens = list(tokenize_text(sent))
            encoded_words = self.cae.encode(tokens)
            encoded_sent = self._combine_fn(encoded_words)
            output.append(torch.Tensor(encoded_sent))
        output = torch.stack(output)
        return output

    def decode(self, vectors):
        if self._mode == 'concat':

            vec_size = self.cae.output_size

            output = []
            for vec in vectors:

                viewed_vec = vec.view(-1, vec_size)

                # Find index of first padding vector
                for index, v in enumerate(viewed_vec):
                    if v.abs().sum() == 0:
                        break
                else:
                    index = viewed_vec.size(0)

                out = self.cae.decode(
                    viewed_vec[:index]
                )

                output.append(out)

            return output

        elif self._mode == 'mean':
            raise ValueError('decode is only defined for mode="concat"')
        else:
            self._unexpected_mode()



from typing import Dict, Optional
from functools import partial
import multiprocessing as mp

import numpy as np
import pandas as pd
from lightwood.helpers.parallelism import get_nr_procs

from type_infer.dtype import dtype
from lightwood.api.types import TimeseriesSettings, PredictionArguments
from lightwood.helpers.log import log


def transform_timeseries(
        data: pd.DataFrame, dtype_dict: Dict[str, str],
        timeseries_settings: TimeseriesSettings, target: str, mode: str,
        pred_args: Optional[PredictionArguments] = None
) -> pd.DataFrame:
    """
    Block that transforms the dataframe of a time series task to a convenient format for use in posterior phases like model training.
    
    The main transformations performed by this block are:
      - Type casting (e.g. to numerical for `order_by` column).
      - Windowing functions for historical context based on `TimeseriesSettings.window` parameter.
      - Explicitly add target columns according to the `TimeseriesSettings.horizon` parameter.
      - Flag all rows that are "predictable" based on all `TimeseriesSettings`.
      - Plus, handle all logic for the streaming use case (where forecasts are only emitted for the last observed data point).
    
    :param data: Dataframe with data to transform.
    :param dtype_dict: Dictionary with the types of each column.
    :param timeseries_settings: A `TimeseriesSettings` object.
    :param target: The name of the target column to forecast.
    :param mode: Either "train" or "predict", depending on what phase is calling this procedure.
    :param pred_args: Optional prediction arguments to control the transformation process.
    
    :return: A dataframe with all the transformations applied.
    """  # noqa

    pred_args = PredictionArguments() if pred_args is None else pred_args
    tss = timeseries_settings
    gb_arr = tss.group_by if tss.group_by is not None else []
    oby = tss.order_by
    window = tss.window
    oby_col = tss.order_by

    if tss.use_previous_target and target not in data.columns:
        raise Exception(f"Cannot transform. Missing historical values for target column {target} (`use_previous_target` is set to True).")  # noqa

    for hcol in tss.historical_columns:
        if hcol not in data.columns or data[hcol].isna().any():
            raise Exception(f"Cannot transform. Missing values in historical column {hcol}.")

    # initial stable sort and per-partition deduplication TODO: slowish, add a top-level param to disable if needed
    data = data.sort_values(by=oby_col, kind='mergesort')
    data = data.drop_duplicates(subset=[oby_col, *gb_arr], keep='first')

    # pass seconds to timestamps according to each group's inferred freq, and force this freq on index
    grouped = data.groupby(by=tss.group_by) if tss.group_by else data.groupby(lambda x: True)
    reindexed = []
    # TODO: introduce MP here
    for name, group in grouped:
        name = name if tss.group_by and len(tss.group_by) > 1 else (name, )  # guaranteed tuple type
        if group.shape[0] > 0:
            if group[tss.order_by].value_counts().max() > 1 and group.shape[0] > 1:
                raise Exception(f"Partition is not valid, faulty group {name}. Please make sure you group by a set of columns that ensures unique measurements for each grouping through time.")  # noqa

            index = pd.to_datetime(group[oby_col], unit='s', utc=True)
            group.index = pd.date_range(start=index.iloc[0], end=index.iloc[-1], periods=len(group))
            resampled = group
            group['__mdb_inferred_freq'] = None
            if len(group) > 2:
                freq = pd.infer_freq(group.index)
                if freq is not None:
                    group['__mdb_inferred_freq'] = freq  # sets constant column because pd.concat forgets freq (see: https://github.com/pandas-dev/pandas/issues/3232)  # noqa
                    resampled = group.resample(freq).first()
            reindexed.append(resampled)

    original_df = pd.concat(reindexed).sort_values(by='__mdb_original_index')

    if '__mdb_forecast_offset' in original_df.columns:
        """ This special column can be either None or an integer. If this column is passed, then the TS transformation will react to the values within:

        * If all rows = `None`, proceed as usual. This ends up generating one HORIZON-length forecast for each row in the DF.
        * If all rows have the same value `N <= 0`, then cutoff the dataframe latest `-N` rows after TS shaping and prime the DF (with `__make_predictions` column) so that a forecast is generated only for the last row (thus more efficient). This enables `WHERE T = LATEST - K` with `0 <= K < WINDOW` syntax upstream in MindsDB.
        * If all rows have the same value `N = 1`, then activate streaming inference mode where a single forecast will be emitted for the timestamp inferred by the `_ts_infer_next_row` method. This enables the (already supported) `WHERE T > LATEST` syntax.
        """  # noqa
        index = original_df[~original_df['__mdb_forecast_offset'].isin([None])]  # trigger if col is constant & != None
        offset_available = index.shape[0] == len(original_df) and \
            original_df['__mdb_forecast_offset'].unique().tolist() != [None]
        if offset_available:
            offset = min(int(original_df['__mdb_forecast_offset'].unique()[0]), 1)
        else:
            offset = 0
        cutoff_mode = offset_available and offset == 1
    else:
        offset_available = False
        offset = 0
        cutoff_mode = False

    if '__mdb_forecast_offset' in original_df.columns or cutoff_mode:
        original_df['__make_predictions'] = True
        original_df['original_index'] = np.arange(len(original_df))
    else:
        original_df['__make_predictions'] = False
        original_df['original_index'] = None

    secondary_type_dict = {}
    if dtype_dict[oby] in (dtype.date, dtype.integer, dtype.float):
        secondary_type_dict[oby] = dtype_dict[oby]

    original_df[f'__mdb_original_{oby}'] = original_df[oby]
    original_df = _ts_to_obj(original_df, [oby] + tss.historical_columns)
    group_lengths = []
    if len(gb_arr) > 0:
        df_arr = []
        for _, df in original_df.groupby(gb_arr):
            df_arr.append(df.sort_values(by=oby))
            group_lengths.append(len(df))
    else:
        df_arr = [original_df.sort_values(by=oby)]
        group_lengths.append(len(original_df))

    n_groups = len(df_arr)
    for i, subdf in enumerate(df_arr):
        if '__mdb_forecast_offset' in subdf.columns and mode == 'predict':
            if cutoff_mode:
                df_arr[i] = _ts_infer_next_row(subdf, oby)
                make_preds = [False for _ in range(max(0, len(df_arr[i]) - 1))] + [True]
            elif offset_available:
                # truncate to forecast up until some len(df) + offset (which is <= 0)
                new_index = df_arr[i].index[:len(df_arr[i].index) + offset]
                make_preds = [False for _ in range(max(0, len(new_index) - 1))] + [True]
                df_arr[i] = df_arr[i].loc[new_index]
            else:
                if pred_args.force_ts_infer:
                    df_arr[i] = _ts_infer_next_row(subdf, oby)  # force-infer out-of-sample forecast in default mode
                make_preds = [True for _ in range(len(df_arr[i]))]
            df_arr[i]['__make_predictions'] = make_preds

    if len(df_arr) > 1 and len(original_df) > 5000:
        # @TODO: restore possibility to override this with args
        biggest_sub_df = df_arr[np.argmax(group_lengths)]
        nr_procs = min(get_nr_procs(biggest_sub_df), len(df_arr))
        log.info(f'Using {nr_procs} processes to reshape.')
        with mp.Pool(processes=nr_procs) as pool:
            df_arr = pool.map(
                partial(_ts_add_previous_rows, order_cols=[oby] + tss.historical_columns, window=window),
                df_arr
            )

            df_arr = pool.map(
                partial(_ts_add_future_target, target=target, horizon=tss.horizon,
                        data_dtype=tss.target_type, mode=mode),
                df_arr
            )

            if tss.use_previous_target:
                df_arr = pool.map(
                    partial(_ts_add_previous_target, target=target, window=tss.window),
                    df_arr
                )
    else:
        for i in range(n_groups):
            df_arr[i] = _ts_add_previous_rows(df_arr[i],
                                              order_cols=[oby] + tss.historical_columns, window=window)
            df_arr[i] = _ts_add_future_target(df_arr[i], target=target, horizon=tss.horizon,
                                              data_dtype=tss.target_type, mode=mode)
            if tss.use_previous_target:
                df_arr[i] = _ts_add_previous_target(df_arr[i], target=target, window=tss.window)

    combined_df = pd.concat(df_arr)

    if '__mdb_forecast_offset' in combined_df.columns:
        combined_df = pd.DataFrame(combined_df[combined_df['__make_predictions']])  # filters by True only
        del combined_df['__make_predictions']

    if not cutoff_mode and any([i < tss.window for i in group_lengths]):
        if tss.allow_incomplete_history:
            log.warning("Forecasting with incomplete historical context, predictions might be subpar")
        else:
            raise Exception(f'Not enough historical context to make a timeseries prediction (`allow_incomplete_history` is set to False). Please provide a number of rows greater or equal to the window size - currently (number_rows, window_size) = ({min(group_lengths)}, {tss.window}). If you can\'t get enough rows, consider lowering your window size. If you want to force timeseries predictions lacking historical context please set the `allow_incomplete_history` timeseries setting to `True`, but this might lead to subpar predictions depending on the mixer.') # noqa

    if n_groups > 1:
        df_gb_list = list(combined_df.groupby(tss.group_by))
        df_gb_map = {}
        for gb, df in df_gb_list:
            df_gb_map['_' + '_'.join(str(gb))] = df

    del combined_df['original_index']

    return combined_df


def _ts_infer_next_row(df: pd.DataFrame, ob: str) -> pd.DataFrame:
    """
    Adds an inferred next row for streaming mode purposes.

    :param df: dataframe from which next row is inferred.
    :param ob: `order_by` column.

    :return: Modified `df` with the inferred row appended to it.
    """
    original_index = df.index.copy()
    start = original_index.min()
    new_index = pd.date_range(start=start, periods=len(original_index) + 1, freq=df['__mdb_inferred_freq'].iloc[0])
    last_row = df.iloc[[-1]].copy()
    last_row['__mdb_ts_inferred'] = True

    if df.shape[0] > 1:
        butlast_row = df.iloc[[-2]]
        delta = (last_row[ob].values - butlast_row[ob].values).flatten()[0]
    else:
        delta = 1

    last_row[ob] += delta
    new_df = pd.concat([df, last_row], ignore_index=True)
    new_df.index = pd.DatetimeIndex(new_index)
    return new_df


def _ts_to_obj(df: pd.DataFrame, historical_columns: list) -> pd.DataFrame:
    """
    Casts all historical columns in a dataframe to `object` type.

    :param df: Input dataframe
    :param historical_columns: Historical columns to type cast

    :return: Dataframe with `object`-typed historical columns
    """
    for hist_col in historical_columns:
        df.loc[:, hist_col] = df[hist_col].astype(object)
    return df


def _ts_add_previous_rows(df: pd.DataFrame, order_cols: list, window: int) -> pd.DataFrame:
    """
    Adds previous rows (as determined by `TimeseriesSettings.window`) into the cells of the `order_by` column.

    :param df: Input dataframe.
    :param order_cols: `order_by` column and other columns flagged as `historical`.
    :param window: value of `TimeseriesSettings.window` parameter.
    
    :return: Dataframe with all `order_cols` modified so that their values are now arrays of historical context.
    """  # noqa
    for order_col in order_cols:
        new_vals = np.zeros((len(df), window))
        for i in range(window, 0, -1):
            new_vals[:, i - 1] = df[order_col].shift(window - i).values

        new_vals = np.nan_to_num(new_vals, nan=0.0)
        for i in range(len(df)):
            df.at[df.index[i], order_col] = new_vals[i, :]

    return df


def _ts_add_previous_target(df: pd.DataFrame, target: str, window: int) -> pd.DataFrame:
    """
    Adds previous rows (as determined by `TimeseriesSettings.window`) into the cells of the target column.

    :param df: Input dataframe.
    :param target: target column name.
    :param window: value of `TimeseriesSettings.window` parameter.

    :return: Dataframe with new `__mdb_ts_previous_{target}` column that contains historical target context.
    """  # noqa
    if target not in df:
        return df
    previous_target_values = list(df[target])
    del previous_target_values[-1]
    previous_target_values = [None] + previous_target_values

    previous_target_values_arr = []
    for i in range(len(previous_target_values)):
        prev_vals = previous_target_values[max(i - window, 0):i + 1]
        arr = [None] * (window - len(prev_vals))
        arr.extend(prev_vals)
        previous_target_values_arr.append(arr)

    df[f'__mdb_ts_previous_{target}'] = previous_target_values_arr
    return df


def _ts_add_future_target(df, target, horizon, data_dtype, mode):
    """
    Adds as many columns to the input dataframe as the forecasting horizon asks for (as determined by `TimeseriesSettings.horizon`).

    :param df: Input dataframe.
    :param target: target column name.
    :param horizon: value of `TimeseriesSettings.horizon` parameter.
    :param data_dtype: dictionary with types of all input columns
    :param mode: either "train" or "predict". `Train` will drop rows with incomplet target info. `Predict` has no effect, for now.

    :return: Dataframe with new `{target}_timestep_{i}'` columns that contains target labels at timestep `i` of a total `TimeseriesSettings.horizon`.
    """  # noqa
    if target not in df:
        return df
    if data_dtype in (dtype.integer, dtype.float, dtype.num_array, dtype.num_tsarray):
        df[target] = df[target].astype(float)

    for timestep_index in range(1, horizon):
        next_target_value_arr = list(df[target])
        for del_index in range(0, min(timestep_index, len(next_target_value_arr))):
            del next_target_value_arr[0]
            next_target_value_arr.append(None)
        col_name = f'{target}_timestep_{timestep_index}'
        df[col_name] = next_target_value_arr
        df[col_name] = df[col_name].fillna(value=np.nan)

    return df



# Phases
from lightwood.analysis.analyze import model_analyzer
from lightwood.analysis.explain import explain

# Base blocks
from lightwood.analysis.base import BaseAnalysisBlock
from lightwood.analysis.nc.calibrate import ICP
from lightwood.analysis.helpers.acc_stats import AccStats
from lightwood.analysis.helpers.conf_stats import ConfStats
from lightwood.analysis.nn_conf.temp_scale import TempScaler
from lightwood.analysis.helpers.feature_importance import PermutationFeatureImportance

# Blocks with extra requirements
try:
    from lightwood.analysis.helpers.shap import ShapleyValues
    from lightwood.analysis.helpers.pyod import PyOD
except Exception:
    ShapleyValues = None
    PyOD = None


__all__ = ['model_analyzer', 'explain', 'BaseAnalysisBlock', 'TempScaler', 'PyOD',
           'ICP', 'AccStats', 'ConfStats', 'PermutationFeatureImportance', 'ShapleyValues']



from copy import deepcopy as dc
from typing import Dict, List, Iterable, Tuple

import torch
import numpy as np
from scipy.special import softmax

from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.constants import _UNCOMMON_WORD
from lightwood.helpers.log import log


class BinaryEncoder(BaseEncoder):
    """
   Creates a one-hot-encoding for binary class data. Assume two arbitrary categories :math:`A` and :math:`B`; representation for them will be as such:

    .. math::

       A &= [1, 0] \\
       B &= [0, 1]

    This encoder is a specialized case of one-hot encoding (OHE); unknown categories are explicitly handled as [0, 0]. Unknowns may only be reported if the input row value is NULL (or python `None` type) or if new data, after the encoder is prepared, has examples outside the feature map.

    When data is typed with Lightwood, this class is only deployed if an input data type is explicitly recognized as binary (i.e. the column has only 2 unique values like True/False). If future data shows a new category (thus the data is no longer truly binary), this encoder will no longer be appropriate unless you are comfortable mapping ALL new classes as [0, 0].

    An encoder can represent a feature column or target column; in this case it represents a target, `is_target` is `True`, and `target_weights`. The `target_weights` parameter enables users to specify how heavily each class should be weighted within a mixer - useful in imbalanced classes. 

    By default, `dataprep_ml.StatisticalAnalysis` phase will provide `target_weights` as the relative fraction of each class in the data which is important for imbalanced populations; for example, suppose there is a 80/20 imbalanced representation across 3 different classes - `target_weights` will be a vector as such::

    target_weights = {"class1": 0.8, "class2": 0.2}

    Users should note that models will be presented with the inverse of the target weights, `inv_target_weights`, which will perform the 1/target_value_per_class operation. **This means large values will result in small weights for the model**.
    """ # noqa
    def __init__(
        self,
        is_target: bool = False,
        target_weights: Dict[str, float] = None,
        handle_unknown: str = 'use_encoded_value'
    ):
        super().__init__(is_target)
        """
        :param is_target: Whether encoder featurizes target column
        :param target_weights: Percentage of total population represented by each category (from [0, 1]), as a dictionary.
        :param handle_unknown: if set to `use_encoded_value`, will assign all classes with index greater than 1 to a special UNKNOWN index. This doesn't affect the encoded representation of shape (B, 2). During decoding, any unknown or otherwise known but "out-of-bounds" word will be decoded back to the lightwood unknown category token. If this argument is set to `error`, the encoder will raise an error while preparing if there are more than two observed classes.
        """  # noqa

        self.map = {}  # category name -> index
        self.rev_map = {}  # index -> category name
        self.output_size = 2
        self.encoder_class_type = str
        self.handle_unknown = handle_unknown

        # Weight-balance info if encoder represents target
        self.target_weights = None
        self.index_weights = None
        if self.is_target:
            self.target_weights = dc(target_weights)

    def prepare(self, priming_data: Iterable[str]):
        """
        Given priming data, create a map/inverse-map corresponding category name to index (and vice versa).

        :param priming_data: Binary data to encode
        """  # noqa
        if self.is_prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        unq_cats = np.unique([i for i in priming_data if i is not None]).tolist()

        self.map = {cat: indx for indx, cat in enumerate(unq_cats)}
        self.rev_map = {indx: cat for cat, indx in self.map.items()}

        # Enforce only binary; map must have exactly 2 classes.
        if len(self.map) > 2 and self.handle_unknown == 'use_encoded_value':
            log.warning('Warning: dtype for binary encoder has > 2 classes. Extra classes will be pointed to an invalid token. Try overriding this encoder with a multi-class categorical encoder, otherwise performance may not be optimal.')  # noqa
            log.warning(f'Observed classes are: {self.map}.')
        elif self.handle_unknown == 'error':
            raise ValueError(f'Data has > 2 classes and encoder is in strict mode. Aborting. All classes are: {self.map}.')  # noqa

        # For target-only, report on relative weights of classes
        if self.is_target:

            self.index_weights = torch.ones(self.output_size)  # Equally wt. both classes

            # If target weights provided, weight by inverse
            if self.target_weights is not None:

                # Check target weights properly specified
                if sum([np.isclose(i, 0) for i in self.target_weights.values()]) > 0:
                    raise ValueError('Target weights cannot be 0')

                # Ensure all classes are specified in the weights criteria
                assert(set(self.target_weights.keys()) == set(self.map.keys()))

                for cat in self.map.keys():
                    self.index_weights[self.map[cat]] = self.target_weights[cat]

        self.is_prepared = True

    def encode(self, column_data: Iterable[str]) -> torch.Tensor:
        """
        Encodes categories as OHE binary. Unknown/unrecognized classes return [0,0].

        :param column_data: Pre-processed data to encode
        :returns Encoded data of form :math:`N_{rows} x 2`
        """  # noqa
        if not self.is_prepared:
            raise Exception(
                'You need to call "prepare" before calling "encode" or "decode".'
            )

        ret = torch.zeros(size=(len(column_data), self.output_size))

        for idx, word in enumerate(column_data):
            index = self.map.get(word, None)

            if index is None or index >= self.output_size:
                pass  # any unknown value is ignored
            else:
                ret[idx, index] = 1

        return torch.Tensor(ret)

    def decode(self, encoded_data: torch.Tensor):
        """
        Given encoded data, return in form of original category labels.
        The input to `decode` makes no presumption on whether the data is already in OHE form OR not, as it some models may output a set of probabilities of weights assigned to each class. The decoded value will always be the argmax of such a vector.

        In the case that the vector is all 0s, the output is decoded as "UNKNOWN"

        :param encoded_data: the output of a mixer model

        :returns: Decoded values for each data point
        """  # noqa
        encoded_data_list = encoded_data.tolist()
        ret = []

        for vector in encoded_data_list:
            if not np.any(vector):  # Vector of all 0s -> unknown category
                ret.append(_UNCOMMON_WORD)
            else:
                idx = np.argmax(vector)
                if idx >= self.output_size:
                    ret.append(_UNCOMMON_WORD)  # known, but not either of the supported categories
                else:
                    ret.append(self.rev_map[idx])

        return ret

    def decode_probabilities(self, encoded_data: torch.Tensor) -> Tuple[List[str], List[List[float]], Dict[int, str]]:
        """
        Provides decoded answers, as well as a probability assignment to each data point.

        :param encoded_data: the output of a mixer model

        :returns: Decoded values for each data point, Probability vector for each category, and the reverse map of dimension to category name
        """  # noqa
        encoded_data_list = encoded_data.tolist()
        ret = []
        probs = []

        for vector in encoded_data_list:
            if not np.any(vector):  # Vector of all 0s -> unknown category
                ret.append(_UNCOMMON_WORD)
            else:
                ret.append(self.rev_map[np.argmax(vector)])

            probs.append(self._norm_vec(vector))

        return ret, probs, self.rev_map

    @staticmethod
    def _norm_vec(vec: List[float]):
        """
        Given a vector, normalizes so that the sum of elements is 1, using softmax.

        :param vec: Assigned weights for each category
        """  # noqa
        return softmax(vec)



__title__ = 'lightwood'
__package_name__ = 'lightwood'
__version__ = '24.3.3.1'
__description__ = "Lightwood is a toolkit for automatic machine learning model building"
__email__ = "community@mindsdb.com"
__author__ = 'MindsDB Inc'
__github__ = 'https://github.com/mindsdb/lightwood'
__pypi__ = 'https://pypi.org/project/lightwood'
__license__ = 'GPL-3.0'
__copyright__ = 'Copyright 2019- mindsdb'



from typing import Dict, Tuple, List

import numpy as np
import pandas as pd
from type_infer.dtype import dtype

from lightwood.api.types import TimeseriesSettings
from lightwood.helpers.ts import get_ts_groups, get_delta, Differencer
from lightwood.encoder.time_series.helpers.common import generate_target_group_normalizers


def timeseries_analyzer(data: Dict[str, pd.DataFrame], dtype_dict: Dict[str, str],
                        timeseries_settings: TimeseriesSettings, target: str) -> Dict:
    """
    This module analyzes (pre-processed) time series data and stores a few useful insights used in the rest of Lightwood's pipeline.
    
    :param data: dictionary with the dataset split into train, val, test subsets. 
    :param dtype_dict: dictionary with inferred types for every column.
    :param timeseries_settings: A `TimeseriesSettings` object. For more details, check `lightwood.types.TimeseriesSettings`.
    :param target: name of the target column.
    
    The following things are extracted from each time series inside the dataset:
      - group_combinations: all observed combinations of values for the set of `group_by` columns. The length of this list determines how many time series are in the data.
      - deltas: inferred sampling interval 
      - ts_naive_residuals: Residuals obtained from the data by a naive forecaster that repeats the last-seen value. 
      - ts_naive_mae: Mean residual value obtained from the data by a naive forecaster that repeats the last-seen value.
      - target_normalizers: objects that may normalize the data within any given time series for effective learning. See `lightwood.encoder.time_series.helpers.common` for available choices.
    
    :return: Dictionary with the aforementioned insights and the `TimeseriesSettings` object for future references.
    """  # noqa
    tss = timeseries_settings
    groups = get_ts_groups(data['train'], tss)
    deltas, periods, freqs = get_delta(data['train'], tss)

    normalizers = generate_target_group_normalizers(data['train'], target, dtype_dict, tss)

    if dtype_dict[target] in (dtype.integer, dtype.float, dtype.num_tsarray):
        naive_forecast_residuals, scale_factor = get_grouped_naive_residuals(data['dev'], target, tss)
        differencers = get_differencers(data['train'], target, tss.group_by)
    else:
        naive_forecast_residuals, scale_factor = {}, {}
        differencers = {}

    return {'target_normalizers': normalizers,
            'deltas': deltas,
            'tss': tss,
            'group_combinations': groups,
            'ts_naive_residuals': naive_forecast_residuals,
            'ts_naive_mae': scale_factor,
            'periods': periods,
            'sample_freqs': freqs,
            'stl_transforms': {},  # TODO: remove, or provide from outside as user perhaps
            'differencers': differencers
            }


def get_naive_residuals(target_data: pd.DataFrame, m: int = 1) -> Tuple[List, float]:
    """
    Computes forecasting residuals for the naive method (forecasts for time `t` is the value observed at `t-1`).
    Useful for computing MASE forecasting error.

    As per arxiv.org/abs/2203.10716, we resort to a constant forecast based on the last-seen measurement across the entire horizon.
    By following the original measure, the naive forecaster would have the advantage of knowing the actual values whereas the predictor would not.

    Note: method assumes predictions are all for the same group combination. For a dataframe that contains multiple
     series, use `get_grouped_naive_resiudals`.

    :param target_data: observed time series targets
    :param m: season length. the naive forecasts will be the m-th previously seen value for each series

    :return: (list of naive residuals, average residual value)
    """  # noqa
    # @TODO: support categorical series as well
    residuals = np.abs(target_data.values[1:] - target_data.values[0]).flatten()
    scale_factor = np.average(residuals)
    return residuals.tolist(), scale_factor


def get_grouped_naive_residuals(
        info: pd.DataFrame,
        target: str,
        tss: TimeseriesSettings
) -> Tuple[Dict, Dict]:
    """
    Wraps `get_naive_residuals` for a dataframe with multiple co-existing time series.
    """  # noqa
    group_residuals = {}
    group_scale_factors = {}
    grouped = info.groupby(by=tss.group_by) if tss.group_by else info.groupby(lambda x: '__default')
    for group, subset in grouped:
        if subset.shape[0] > 1:
            residuals, scale_factor = get_naive_residuals(subset[target])  # @TODO: pass m once we handle seasonality
            group_residuals[group] = residuals
            group_scale_factors[group] = scale_factor
    return group_residuals, group_scale_factors


def get_differencers(data: pd.DataFrame, target: str, group_cols: List):
    differencers = {}
    grouped = data.groupby(by=group_cols) if group_cols else data.groupby(lambda x: True)
    for group, subset in grouped:
        differencer = Differencer()
        differencer.fit(subset[target].values)
        differencers[group] = differencer
    return differencers



from typing import List, Tuple, Iterable
import torch
from lightwood.encoder.image.helpers.img_to_vec import Img2Vec
from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.log import log

try:
    import torchvision.transforms as transforms
    from PIL import Image
except ModuleNotFoundError:
    log.info("No torchvision/pillow detected, image encoder not supported")


class Img2VecEncoder(BaseEncoder):
    """
    Generates encoded representations for images using a pre-trained deep neural network. Inputs must be str-based location of the data.

    Without user-specified details, all input images are rescaled to a standard size of 224x224, and normalized using the mean and standard deviation of the ImageNet dataset (as it was used to train the underlying NN).
    
    This encoder currently does not support a `decode()` call; models with an image output will not work. 
    
    For more information about the neural network this encoder uses, refer to the `lightwood.encoder.image.helpers.img_to_vec.Img2Vec`.
    """  # noqa

    is_trainable_encoder: bool = True

    def __init__(
        self,
        stop_after: float = 3600,
        is_target: bool = False,
        scale: Tuple[int, int] = (224, 224),
        mean: List[float] = [0.485, 0.456, 0.406],
        std: List[float] = [0.229, 0.224, 0.225],
        device: str = '',
    ):
        """
        :param stop_after: time budget, in seconds. 
        :param is_target: Whether encoder represents target or not
        :param scale: Resize scale of image (x, y)
        :param mean: Mean of pixel values
        :param std: Standard deviation of pixel values
        :param device: Name of the device that get_device_from_name will attempt to use
        """  # noqa
        assert not is_target
        super().__init__(is_target)

        self.is_prepared = False
        self.scale = scale
        self.mean = mean
        self.std = std

        self._scaler = transforms.Resize(scale)
        self._normalize = transforms.Normalize(mean=self.mean, std=self.std)
        self._to_tensor = transforms.ToTensor()
        self._img_to_tensor = transforms.Compose([
            self._scaler,
            self._to_tensor,
            self._normalize
        ])
        self.stop_after = stop_after

        self.device_name = device

        # pil_logger = logging.getLogger('PIL')  # noqa
        # pil_logger.setLevel(logging.ERROR)  # noqa

    def prepare(self, train_priming_data: Iterable[str], dev_priming_data: Iterable[str]):
        # @TODO: finetune here? depending on time aim
        """
        Sets an `Img2Vec` object (model) and sets the expected size for encoded representations.
        """
        if self.is_prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        self.model = Img2Vec(device=self.device_name)
        self.device = self.model.device
        self.device_name = f'{self.device}'
        self.output_size = self.model.output_size
        self.is_prepared = True

    def to(self, device, available_devices=1):
        """
        Changes device of model to support CPU/GPU

        :param device: will move the model to this device.
        :param available_devices: all available devices as reported by lightwood.

        :return: same object but moved to the target device.
        """
        self.model.to(device, available_devices=1)
        return self

    def encode(self, images: List[str]) -> torch.Tensor:
        """
        Creates encodings for a list of images; each image is referenced by a filepath or url.

        :param images: list of images, each image is a path to a file or a url.
        :return: a torch.floatTensor
        """
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        img_tensors = [self._img_to_tensor(
            Image.open(img_path)
        ) for img_path in images]
        vec_arr = []

        self.model.eval()
        with torch.no_grad():
            for img_tensor in img_tensors:
                vec = self.model(img_tensor.unsqueeze(0), batch=False)
                vec_arr.append(vec)
        return torch.stack(vec_arr).to('cpu')

    def decode(self, encoded_values_tensor: torch.Tensor):
        """ Currently not supported """
        raise Exception('This encoder is not bi-directional')



from typing import List, Optional

import pandas as pd

from lightwood.mixer.base import BaseMixer
from lightwood.data.encoded_ds import EncodedDs
from lightwood.api.types import PredictionArguments, SubmodelData


class BaseEnsemble:
    """
    Base class for all ensembles.

    Ensembles wrap sets of Lightwood mixers, with the objective of generating better predictions based on the output of each mixer. 
    
    There are two important methods for any ensemble to work:
        1. `__init__()` should prepare all mixers and internal ensemble logic.
        2. `__call__()` applies any aggregation rules to generate final predictions based on the output of each mixer. 

    Class Attributes:
    - mixers: List of mixers the ensemble will use.
    - supports_proba: For classification tasks, whether the ensemble supports yielding per-class scores rather than only returning the predicted label. 

    NOTE: this ensemble is not functional. Do not use it when generating custom JsonAI objects, as the learning process will fail.
    """  # noqa
    data: EncodedDs
    mixers: List[BaseMixer]
    best_index: int  # @TODO: maybe only applicable to BestOf
    supports_proba: bool
    prepared: bool
    submodel_data: List[SubmodelData]

    def __init__(self, target, mixers: List[BaseMixer], data: EncodedDs, fit: Optional[bool] = True) -> None:
        self.data = data
        self.mixers = mixers
        self.best_index = 0
        self.supports_proba = False
        self.submodel_data = []

    def __call__(self, ds: EncodedDs, args: PredictionArguments) -> pd.DataFrame:
        raise NotImplementedError()



from typing import Union, Optional

import torch
import numpy as np
import pandas as pd
from torch.nn.functional import softmax

from type_infer.dtype import dtype


def t_softmax(x, t=1.0, axis=1):
    """ Softmax with temperature scaling """
    return softmax(torch.Tensor(x) / t, dim=axis).numpy()


def clean_df(df, namespace, label_encoders):
    """ Returns cleaned DF for nonconformist calibration """
    # @TODO: reevaluate whether this can be streamlined
    enc = label_encoders
    ns = namespace
    target = ns.target

    if ns.is_classification:
        y = df.pop(target).values
        if enc and isinstance(enc.categories_[0][0], str):
            cats = enc.categories_[0].tolist()
            # the last element is "__mdb_unknown_cat"
            y = np.array([cats.index(i) if i in cats else len(cats) - 1 for i in y])
        y = y.clip(-pow(2, 63), pow(2, 63)).astype(int)
    elif ns.is_multi_ts:
        target_cols = [ns.target] + [f'{ns.target}_timestep_{i}' for i in range(1, ns.tss.horizon)]
        y = np.transpose(np.array([df.pop(col) for col in target_cols]))
    else:
        y = df.pop(target).values
        y = y.astype(float)

    return df, y


def set_conf_range(
        X: pd.DataFrame,
        icp,
        target_type: dtype,
        analysis_info: dict,
        positive_domain: bool = False,
        std_tol: int = 1,
        group: str = '__default',
        significance: Optional[float] = None
):
    """
    Automatically sets confidence level for numerical and categorical tasks.

    :param X: Validation data.
    :param icp: Inductive conformal predictor that sets the confidence level. Either IcpClassifier or IcpRegressor.
    :param target_type: dtype of the target column.
    :param analysis_info:
    :param positive_domain: Flag that indicates whether target is expected to be a positive number.
    :param std_tol: Tolerance for automatic confidence level selection; bigger tolerance means higher confidence, in general.
    :param group: For tasks with multiple different target groups (where each may have a different std_dev), indicates what group is being considered.
    :param significance: Desired confidence level. Can be preset (0 < x <= 0.99)

    :return: set confidence plus predictions regions (for numerical tasks) or pvalues (for categorical tasks).
    """  # noqa
    # numerical
    if target_type in (dtype.integer, dtype.float, dtype.quantity):

        # ICP gets all possible bounds (shape: (B, 2, 99))
        all_ranges = icp.predict(X.values)

        # iterate over confidence levels until spread >= a multiplier of the dataset stddev
        if significance is not None:
            conf = int(100 * (1 - significance))
            return significance, all_ranges[:, :, conf]
        else:
            for tol in [std_tol, std_tol + 1, std_tol + 2]:
                for significance in range(99):
                    ranges = all_ranges[:, :, significance]
                    spread = np.mean(ranges[:, 1] - ranges[:, 0])
                    tolerance = analysis_info['df_target_stddev'][group] * tol

                    if spread <= tolerance:
                        confidence = (99 - significance) / 100
                        if positive_domain:
                            ranges[ranges < 0] = 0
                        return confidence, ranges
            else:
                ranges = all_ranges[:, :, 0]
                if positive_domain:
                    ranges[ranges < 0] = 0
                return 0.9901, ranges

    # time series
    elif target_type in (dtype.num_array, dtype.num_tsarray):
        pass

    # categorical
    elif target_type in (dtype.binary, dtype.categorical, dtype.cat_array, dtype.cat_tsarray):
        pvals = icp.predict(X.values)  # p-values at which each class is included in the predicted set
        conf = get_categorical_conf(pvals)
        return conf, pvals

    # default
    return 0.005, np.zeros((X.shape[0], 2))


def get_numeric_conf_range(
        all_confs: np.ndarray,
        df_target_stddev: dict = {},
        positive_domain: bool = False,
        std_tol: int = 1,
        group: Optional[str] = '__default',
        fixed_conf: float = None
):
    """
    Gets prediction bounds for numerical targets, based on ICP estimation and width tolerance.
    
    :param all_confs: All possible bounds depending on confidence level.
    :param df_target_stddev: Observed train standard deviation for each group target.
    :param positive_domain: Flag that indicates whether target is expected to be a positive number.
    :param std_tol: Tolerance for automatic confidence level selection; bigger tolerance means higher confidence, in general.
    :param group: For tasks with multiple different target groups (where each may have a different std_dev), indicates what group is being considered.
    :param fixed_conf: Pre-determined confidence for the ICP, 0-1 bounded. Can be specified to bypass automatic confidence/bound detection, or to adjust the threshold sensitivity in anomaly detection tasks.
    
    :return: array with confidence for each data instance, along with lower and upper bounds for each prediction.
    """  # noqa

    if fixed_conf is None:
        significances = []
        conf_ranges = []
        std_dev = df_target_stddev[group]
        tolerance = std_dev * std_tol

        for sample_idx in range(all_confs.shape[0]):
            sample = all_confs[sample_idx, :, :]
            for idx in range(sample.shape[1]):
                significance = (99 - idx) / 100
                diff = sample[1, idx] - sample[0, idx]
                if diff <= tolerance:
                    conf_range = list(sample[:, idx])
                    significances.append(significance)
                    conf_ranges.append(conf_range)
                    break
            else:
                significances.append(0.9991)  # default: confident that value falls inside big bounds
                bounds = sample[:, 0]
                sigma = (bounds[1] - bounds[0]) / 4
                conf_range = [bounds[0] - sigma, bounds[1] + sigma]
                conf_ranges.append(conf_range)

        conf_ranges = np.array(conf_ranges)
    else:
        # fixed error rate
        conf = max(0.01, min(1.0, fixed_conf))
        error_rate = 1 - conf
        conf_idx = int(100 * error_rate) - 1
        conf_ranges = all_confs[:, :, conf_idx]
        significances = [conf for _ in range(conf_ranges.shape[0])]

    if positive_domain:
        conf_ranges[conf_ranges < 0] = 0
    return np.array(significances), conf_ranges


def get_ts_conf_range(
        all_confs: np.ndarray,
        df_target_stddev: dict = {},
        positive_domain: bool = False,
        std_tol: int = 1,
        group: Optional[str] = '__default',
        fixed_conf: float = None
):
    all_significances = []
    all_conf_ranges = []
    for timestep in range(all_confs.shape[1]):
        sigs, confs = get_numeric_conf_range(
            all_confs[:, timestep, :, :],
            df_target_stddev,
            positive_domain,
            std_tol,
            group,
            fixed_conf
        )
        all_significances.append(sigs)
        all_conf_ranges.append(confs)

    return np.vstack(all_significances).T, np.stack(all_conf_ranges).swapaxes(0, 1)


def get_categorical_conf(raw_confs: np.ndarray):
    """
    ICP confidence estimation for categorical targets from raw p-values:
        1.0 minus 2nd highest p-value yields confidence for predicted label.
    :param all_confs: p-value for each class per data point
    :return: confidence for each data point
    """
    if len(raw_confs.shape) == 1:
        raw_confs = np.expand_dims(raw_confs, axis=0)
    if raw_confs.shape[-1] == 1:
        # single-class edge case (only happens if predictor sees just one known label at calibration)
        confs = np.clip(raw_confs[:, 0], 0.0001, 0.9999)
    else:
        second_p = np.sort(raw_confs, axis=1)[:, -2]
        confs = np.clip(np.subtract(1, second_p), 0.0001, 0.9999)
    return confs


def get_anomalies(insights: pd.DataFrame, observed_series: Union[pd.Series, list], cooldown: int = 1) -> np.ndarray:
    """
    Simple procedure for unsupervised anomaly detection in time series forecasting. 
    Uses ICP analysis block output so that any true value falling outside of the lower and upper bounds is tagged as anomalous.
      
    :param insights: dataframe with row insights used during the `.explain()` phase of all analysis blocks.  
    :param observed_series: true values from the predicted time series. If empty, no anomalies are flagged.
    :param cooldown: minimum amount of observations (assuming regular sampling frequency) that need to pass between two consecutive anomalies.
    
    :return: np.ndarray of boolean flags, indicating anomalous behavior for each predicted value.
    """  # noqa
    anomalies = []
    counter = 0

    # cast to float (safe, we only call this method if series is numerical)
    try:
        observed_series = [float(value) for value in observed_series]
    except (TypeError, ValueError):
        return [None for _ in observed_series]

    lower_bounds = insights['lower'].tolist()
    upper_bounds = insights['upper'].tolist()

    for (l, u), t in zip(zip(lower_bounds, upper_bounds), observed_series):
        if t is not None:
            if isinstance(l, list):
                anomaly = not (l[0] <= t <= u[0])
            else:
                anomaly = not (l <= t <= u)

            if anomaly and (counter == 0 or counter >= cooldown):
                anomalies.append(anomaly)  # new anomaly event triggers, reset counter
                counter = 1
            elif anomaly and counter < cooldown:
                anomalies.append(False)  # overwrite as not anomalous if still in cooldown
                counter += 1
            else:
                anomalies.append(anomaly)
                counter = 0
        else:
            anomalies.append(None)
            counter += 1

    return np.array(anomalies)



# TODO: type hint the returns

from typing import Dict, List, Optional, Union
import sys

from type_infer.dtype import dtype

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from dataclasses import dataclass
from lightwood.helpers.log import log
from dataclasses_json import dataclass_json
from dataclasses_json.core import _asdict, Json
import json


# See: https://www.python.org/dev/peps/pep-0589/ for how this works
# Not very intuitive but very powerful abstraction, might be useful in other places (@TODO)
class Module(TypedDict):
    """
    Modules are the blocks of code that end up being called from the JSON AI, representing either object instantiations or function calls.

    :param module: Name of the module (function or class name)
    :param args: Argument to pass to the function or constructor
    """ # noqa
    module: str
    args: Dict[str, str]


@dataclass
class TimeseriesSettings:
    """
    For time-series specific problems, more specific treatment of the data is necessary. The following attributes \
        enable time-series tasks to be carried out properly.

    :param is_timeseries: Whether the input data should be treated as time series; if true, this flag is checked in \
        subsequent internal steps to ensure processing is appropriate for time-series data.
    :param order_by: Column by which the data should be ordered.
    :param group_by: Optional list of columns by which the data should be grouped. Each different combination of values\
         for these columns will yield a different series.
    :param window: The temporal horizon (number of rows) that a model intakes to "look back" into when making a\
         prediction, after the rows are ordered by the order_by column and split into groups if applicable.
    :param horizon: The number of points in the future that predictions should be made for, defaults to 1. Once \
        trained, the model will be able to predict up to this many points into the future.
    :param historical_columns: The temporal dynamics of these columns will be used as additional context to train the \
        time series predictor. Note that a non-historical column shall still be used to forecast, but without \
            considering their change through time.
    :param target_type: Automatically inferred dtype of the target (e.g. `dtype.integer`, `dtype.float`).
    :param use_previous_target: Use the previous values of the target column to generate predictions. Defaults to True.
    :param allow_incomplete_history: whether predictions can be made for rows with incomplete historical context (i.e. less than `window` rows have been observed for the datetime that has to be forecasted).
    :param eval_incomplete: whether to consider predictions with incomplete history or target information when evaluating mixer accuracy with the validation dataset.
    :param interval_periods: tuple of tuples with user-provided period lengths for time intervals. Default values will be added for intervals left unspecified. For interval options, check the `timeseries_analyzer.detect_period()` method documentation. e.g.: (('daily', 7),).
    """  # noqa

    is_timeseries: bool
    order_by: str = None
    window: int = None
    group_by: List[str] = None
    use_previous_target: bool = True
    horizon: int = None
    historical_columns: List[str] = None
    target_type: str = (
        ""  # @TODO: is the current setter (outside of initialization) a sane option?
        # @TODO: George: No, I don't think it is, we need to pass this some other way
    )
    allow_incomplete_history: bool = True
    eval_incomplete: bool = False
    interval_periods: tuple = tuple()

    @staticmethod
    def from_dict(obj: Dict):
        """
        Creates a TimeseriesSettings object from python dictionary specifications.

        :param: obj: A python dictionary with the necessary representation for time-series. The only mandatory columns are ``order_by`` and ``window``.

        :returns: A populated ``TimeseriesSettings`` object.
        """ # noqa
        if len(obj) > 0:
            for mandatory_setting, etype in zip(["order_by", "window"], [str, int]):
                if mandatory_setting not in obj:
                    err = f"Missing mandatory timeseries setting: {mandatory_setting}"
                    log.error(err)
                    raise Exception(err)
                if obj[mandatory_setting] and not isinstance(obj[mandatory_setting], etype):
                    err = f"Wrong type for mandatory timeseries setting '{mandatory_setting}': found '{type(obj[mandatory_setting])}', expected '{etype}'"  # noqa
                    log.error(err)
                    raise Exception(err)

            timeseries_settings = TimeseriesSettings(
                is_timeseries=True,
                order_by=obj["order_by"],
                window=obj["window"],
                use_previous_target=obj.get("use_previous_target", True),
                historical_columns=[],
                horizon=obj.get("horizon", 1),
                allow_incomplete_history=obj.get('allow_incomplete_history', True),
                eval_incomplete=obj.get('eval_incomplete', False),
                interval_periods=obj.get('interval_periods', tuple(tuple()))
            )
            for setting in obj:
                timeseries_settings.__setattr__(setting, obj[setting])

        else:
            timeseries_settings = TimeseriesSettings(is_timeseries=False)

        return timeseries_settings

    @staticmethod
    def from_json(data: str):
        """
        Creates a TimeseriesSettings object from JSON specifications via python dictionary.

        :param: data: JSON-config file with necessary Time-series specifications

        :returns: A populated ``TimeseriesSettings`` object.
        """
        return TimeseriesSettings.from_dict(json.loads(data))

    def to_dict(self, encode_json=False) -> Dict[str, Json]:
        """
        Creates a dictionary from ``TimeseriesSettings`` object

        :returns: A python dictionary containing the ``TimeSeriesSettings`` specifications.
        """
        return _asdict(self, encode_json=encode_json)

    def to_json(self) -> Dict[str, Json]:
        """
        Creates JSON config from TimeseriesSettings object
        :returns: The JSON config syntax containing the ``TimeSeriesSettings`` specifications.
        """
        return json.dumps(self.to_dict())


@dataclass
class ProblemDefinition:
    """
    The ``ProblemDefinition`` object indicates details on how the models that predict the target are prepared. \
        The only required specification from a user is the ``target``, which indicates the column within the input \
        data that the user is trying to predict. Within the ``ProblemDefinition``, the user can specify aspects \
        about how long the feature-engineering preparation may take, and nuances about training the models.

    :param target: The name of the target column; this is the column that will be used as the goal of the prediction.
    :param pct_invalid: Number of data points maximally tolerated as invalid/missing/unknown. \
        If the data cleaning process exceeds this number, no subsequent steps will be taken.
    :param unbias_target: all classes are automatically weighted inverse to how often they occur
    :param seconds_per_mixer: Number of seconds maximum to spend PER mixer trained in the list of possible mixers.
    :param seconds_per_encoder: Number of seconds maximum to spend when training an encoder that requires data to \
    learn a representation.
    :param expected_additional_time: Time budget for non-encoder/mixer tasks \
    (ex: data analysis, pre-processing, model ensembling or model analysis)
    :param time_aim: Time budget (in seconds) to train all needed components for the predictive tasks, including \
        encoders and models.
    :param target_weights: indicates to the accuracy functions how much to weight every target class.
    :param positive_domain: For numerical taks, force predictor output to be positive (integer or float).
    :param timeseries_settings: TimeseriesSettings object for time-series tasks, refer to its documentation for \
         available settings.
    :param anomaly_detection: Whether to conduct unsupervised anomaly detection; currently supported only for time-\
        series.
    :param dtype_dict: Mapping of features to types (see `mindsdb.type_infer` for all possible values). This will \
    override the automated type inference results.
    :param ignore_features: The names of the columns the user wishes to ignore in the ML pipeline. Any column name \
        found in this list will be automatically removed from subsequent steps in the ML pipeline.
    :param use_default_analysis: whether default analysis blocks are enabled.
    :param fit_on_all: Whether to fit the model on the held-out validation data. Validation data is strictly \
        used to evaluate how well a model is doing and is NEVER trained. However, in cases where users anticipate new \
            incoming data over time, the user may train the model further using the entire dataset.
    :param strict_mode: crash if an `unstable` block (mixer, encoder, etc.) fails to run.
    :param seed_nr: custom seed to use when generating a predictor from this problem definition.
    """

    target: str
    pct_invalid: float
    unbias_target: bool
    seconds_per_mixer: Optional[int]
    seconds_per_encoder: Optional[int]
    expected_additional_time: Optional[int]
    time_aim: Optional[float]
    target_weights: Optional[List[float]]
    positive_domain: bool
    timeseries_settings: TimeseriesSettings
    anomaly_detection: bool
    use_default_analysis: bool
    embedding_only: bool
    dtype_dict: Optional[dict]
    ignore_features: List[str]
    fit_on_all: bool
    strict_mode: bool
    seed_nr: int

    @staticmethod
    def from_dict(obj: Dict):
        """
        Creates a ProblemDefinition object from a python dictionary with necessary specifications.

        :param obj: A python dictionary with the necessary features for the ``ProblemDefinition`` class.
        Only requires ``target`` to be specified.

        :returns: A populated ``ProblemDefinition`` object.
        """
        target = obj['target']
        pct_invalid = obj.get('pct_invalid', 2)
        unbias_target = obj.get('unbias_target', True)
        seconds_per_mixer = obj.get('seconds_per_mixer', None)
        seconds_per_encoder = obj.get('seconds_per_encoder', None)
        expected_additional_time = obj.get('expected_additional_time', None)

        time_aim = obj.get('time_aim', None)
        if time_aim is not None and time_aim < 10:
            log.warning(f'Your specified time aim of {time_aim} is too short. Setting it to 10 seconds.')

        target_weights = obj.get('target_weights', None)
        positive_domain = obj.get('positive_domain', False)
        dtype_dict = obj.get('dtype_dict', {})
        timeseries_settings = TimeseriesSettings.from_dict(obj.get('timeseries_settings', {}))
        anomaly_detection = obj.get('anomaly_detection', False)
        ignore_features = obj.get('ignore_features', [])
        fit_on_all = obj.get('fit_on_all', True)
        use_default_analysis = obj.get('use_default_analysis', True)
        embedding_only = obj.get('embedding_only', False)
        strict_mode = obj.get('strict_mode', True)
        seed_nr = obj.get('seed_nr', 1)
        problem_definition = ProblemDefinition(
            target=target,
            pct_invalid=pct_invalid,
            unbias_target=unbias_target,
            seconds_per_mixer=seconds_per_mixer,
            seconds_per_encoder=seconds_per_encoder,
            expected_additional_time=expected_additional_time,
            time_aim=time_aim,
            target_weights=target_weights,
            positive_domain=positive_domain,
            timeseries_settings=timeseries_settings,
            anomaly_detection=anomaly_detection,
            dtype_dict=dtype_dict,
            ignore_features=ignore_features,
            use_default_analysis=use_default_analysis,
            embedding_only=embedding_only,
            fit_on_all=fit_on_all,
            strict_mode=strict_mode,
            seed_nr=seed_nr
        )

        return problem_definition

    @staticmethod
    def from_json(data: str):
        """
        Creates a ProblemDefinition Object from JSON config file.

        :param data:

        :returns: A populated ProblemDefinition object.
        """
        return ProblemDefinition.from_dict(json.loads(data))

    def to_dict(self, encode_json=False) -> Dict[str, Json]:
        """
        Creates a python dictionary from the ProblemDefinition object

        :returns: A python dictionary
        """
        return _asdict(self, encode_json=encode_json)

    def to_json(self) -> Dict[str, Json]:
        """
        Creates a JSON config from the ProblemDefinition object

        :returns: TODO
        """
        return json.dumps(self.to_dict())


@dataclass
class JsonAI:
    """
    The JsonAI Class allows users to construct flexible JSON config to specify their ML pipeline. JSON-AI follows a \
    recipe of how to pre-process data, construct features, and train on the target column. To do so, the following \
    specifications are required internally.

    :param encoders: A dictionary of the form: `column_name -> encoder module`
    :param dtype_dict: A dictionary of the form: `column_name -> data type`
    :param dependency_dict: A dictionary of the form: `column_name -> list of columns it depends on`
    :param model: The ensemble and its submodels
    :param problem_definition: The ``ProblemDefinition`` criteria.
    :param identifiers: A dictionary of column names and respective data types that are likely identifiers/IDs within the data. Through the default cleaning process, these are ignored.
    :param cleaner: The Cleaner object represents the pre-processing step on a dataframe. The user can specify custom subroutines, if they choose, on how to handle preprocessing. Alternatively, "None" suggests the default approach in ``dataprep_ml.cleaners``.
    :param splitter: The Splitter object is the method in which the input data is split into training/validation/testing data. For more details, refer to the `dataprep_ml` package documentation.
    :param analyzer: The Analyzer object is used to evaluate how well a model performed on the predictive task.
    :param explainer: The Explainer object deploys explainability tools of interest on a model to indicate how well a model generalizes its predictions.
    :param imputers: A list of objects that will impute missing data on each column. They are called inside the cleaner.
    :param analysis_blocks: The blocks that get used in both analysis and inference inside the analyzer and explainer blocks.
    :param timeseries_transformer: Procedure used to transform any timeseries task dataframe into the format that lightwood expects for the rest of the pipeline.  
    :param timeseries_analyzer: Procedure that extracts key insights from any timeseries in the data (e.g. measurement frequency, target distribution, etc).
    :param accuracy_functions: A list of performance metrics used to evaluate the best mixers.
    """ # noqa

    encoders: Dict[str, Module]
    dtype_dict: Dict[str, dtype]
    dependency_dict: Dict[str, List[str]]
    model: Dict[str, Module]
    problem_definition: ProblemDefinition
    identifiers: Dict[str, str]
    cleaner: Optional[Module] = None
    splitter: Optional[Module] = None
    analyzer: Optional[Module] = None
    explainer: Optional[Module] = None
    imputers: Optional[List[Module]] = None
    analysis_blocks: Optional[List[Module]] = None
    timeseries_transformer: Optional[Module] = None
    timeseries_analyzer: Optional[Module] = None
    accuracy_functions: Optional[List[Union[str, Module]]] = None

    @staticmethod
    def from_dict(obj: Dict):
        """
        Creates a JSON-AI object from dictionary specifications of the JSON-config.
        """
        encoders = obj["encoders"]
        dtype_dict = obj["dtype_dict"]
        dependency_dict = obj["dependency_dict"]
        model = obj["model"]
        problem_definition = ProblemDefinition.from_dict(obj["problem_definition"])
        identifiers = obj["identifiers"]
        cleaner = obj.get("cleaner", None)
        splitter = obj.get("splitter", None)
        analyzer = obj.get("analyzer", None)
        explainer = obj.get("explainer", None)
        imputers = obj.get("imputers", None)
        analysis_blocks = obj.get("analysis_blocks", None)
        timeseries_transformer = obj.get("timeseries_transformer", None)
        timeseries_analyzer = obj.get("timeseries_analyzer", None)
        accuracy_functions = obj.get("accuracy_functions", None)

        json_ai = JsonAI(
            encoders=encoders,
            dtype_dict=dtype_dict,
            dependency_dict=dependency_dict,
            model=model,
            problem_definition=problem_definition,
            identifiers=identifiers,
            cleaner=cleaner,
            splitter=splitter,
            analyzer=analyzer,
            explainer=explainer,
            imputers=imputers,
            analysis_blocks=analysis_blocks,
            timeseries_transformer=timeseries_transformer,
            timeseries_analyzer=timeseries_analyzer,
            accuracy_functions=accuracy_functions,
        )

        return json_ai

    @staticmethod
    def from_json(data: str):
        """ Creates a JSON-AI object from JSON config"""
        return JsonAI.from_dict(json.loads(data))

    def to_dict(self, encode_json=False) -> Dict[str, Json]:
        """
        Creates a python dictionary with necessary modules within the ML pipeline specified from the JSON-AI object.

        :returns: A python dictionary that has the necessary components of the ML pipeline for a given dataset.
        """
        as_dict = _asdict(self, encode_json=encode_json)
        for k in list(as_dict.keys()):
            if k == "features":
                feature_dict = {}
                for name in self.features:
                    feature_dict[name] = self.features[name].to_dict()
                as_dict[k] = feature_dict
            if as_dict[k] is None:
                del as_dict[k]
        return as_dict

    def to_json(self) -> Dict[str, Json]:
        """
        Creates JSON config to represent the necessary modules within the ML pipeline specified from the JSON-AI object.

        :returns: A JSON config that has the necessary components of the ML pipeline for a given dataset.
        """
        return json.dumps(self.to_dict(), indent=4)


@dataclass_json
@dataclass
class SubmodelData:
    name: str
    accuracy: float
    is_best: bool


@dataclass_json
@dataclass
class ModelAnalysis:
    """
    The ``ModelAnalysis`` class stores useful information to describe a model and understand its predictive performance on a validation dataset.
    For each trained ML algorithm, we store:

    :param accuracies: Dictionary with obtained values for each accuracy function (specified in JsonAI)
    :param accuracy_histogram: Dictionary with histograms of reported accuracy by target value.
    :param accuracy_samples: Dictionary with sampled pairs of observed target values and respective predictions.
    :param train_sample_size: Size of the training set (data that parameters are updated on)
    :param test_sample_size: Size of the testing set (explicitly held out)
    :param column_importances: Dictionary with the importance of each column for the model, as estimated by an approach that closely follows a leave-one-covariate-out strategy.
    :param confusion_matrix: A confusion matrix for the validation dataset.
    :param histograms: Histogram for each dataset feature.
    :param dtypes: Inferred data types for each dataset feature.

    """ # noqa

    accuracies: Dict[str, float]
    accuracy_histogram: Dict[str, list]
    accuracy_samples: Dict[str, list]
    train_sample_size: int
    test_sample_size: int
    column_importances: Dict[str, float]
    confusion_matrix: object
    histograms: object
    dtypes: object
    submodel_data: List[SubmodelData]


@dataclass
class PredictionArguments:
    """
    This class contains all possible arguments that can be passed to a Lightwood predictor at inference time.
    On each predict call, all arguments included in a parameter dictionary will update the respective fields
    in the `PredictionArguments` instance that the predictor will have.
    
    :param predict_proba: triggers (where supported) predictions in raw probability output form. I.e. for classifiers,
    instead of returning only the predicted class, the output additionally includes the assigned probability for
    each class.   
    :param all_mixers: forces an ensemble to return predictions emitted by all its internal mixers.
    :param mixer_weights: a list with coefficients that are normalized into 0-1 bounded scores to mix the output of all mixers available to a compatible ensemble (e.g. [0.5, 0.5] for an ensemble with two mixers would yield the mean prediction). Can be used with WeightedMeanEnsemble, StackedEnsemble or TsStackedEnsemble.
    :param fixed_confidence: Used in the ICP analyzer module, specifies an `alpha` fixed confidence so that predictions, in average, are correct `alpha` percent of the time. For unsupervised anomaly detection, this also translates into the expected error rate. Bounded between 0.01 and 0.99 (respectively implies wider and tighter bounds, all other parameters being equal).
    :param anomaly_cooldown: Sets the minimum amount of timesteps between consecutive firings of the the anomaly detector.
    :param simple_ts_bounds: in forecasting contexts, enabling this parameter disables the usual conformal-based bounds (with Bonferroni correction) and resorts to a simpler way of scaling bounds through the horizon based on the uncertainty estimation for the first value in the forecast (see helpers.ts.add_tn_num_conf_bounds for the implementation).
    :param anomaly_cooldown: Sets the minimum amount of timesteps between consecutive firings of the the anomaly \
        detector.
    :param time_format: For time series predictors. If set to `infer`, predicted `order_by` timestamps will be formatted back to the original dataset's `order_by` format. Any other string value will be used as a formatting string, unless empty (''), which disables the feature (this is the default behavior).
    :param force_ts_infer: For time series predictors. If set to `true`, an additional row will be produced per each group in the input DF, corresponding to an out-of-sample forecast w.r.t. to the input timestamps.
    """  # noqa

    predict_proba: bool = True
    all_mixers: bool = False
    mixer_weights: list = None
    fixed_confidence: Union[int, float, None] = None
    anomaly_cooldown: int = 1
    forecast_offset: int = 0
    simple_ts_bounds: bool = False
    time_format: str = ''
    force_ts_infer: bool = False
    return_embedding: bool = False

    @staticmethod
    def from_dict(obj: Dict):
        """
        Creates a ``PredictionArguments`` object from a python dictionary with necessary specifications.

        :param obj: A python dictionary with the necessary features for the ``PredictionArguments`` class.

        :returns: A populated ``PredictionArguments`` object.
        """

        # maybe this should be stateful instead, and save the latest used value for each field?
        predict_proba = obj.get('predict_proba', PredictionArguments.predict_proba)
        all_mixers = obj.get('all_mixers', PredictionArguments.all_mixers)
        mixer_weights = obj.get('mixer_weights', PredictionArguments.mixer_weights)
        fixed_confidence = obj.get('fixed_confidence', PredictionArguments.fixed_confidence)
        anomaly_cooldown = obj.get('anomaly_cooldown', PredictionArguments.anomaly_cooldown)
        forecast_offset = obj.get('forecast_offset', PredictionArguments.forecast_offset)
        simple_ts_bounds = obj.get('simple_ts_bounds', PredictionArguments.simple_ts_bounds)
        time_format = obj.get('time_format', PredictionArguments.time_format)
        force_ts_infer = obj.get('force_ts_infer', PredictionArguments.force_ts_infer)
        return_embedding = obj.get('return_embedding', PredictionArguments.return_embedding)

        pred_args = PredictionArguments(
            predict_proba=predict_proba,
            all_mixers=all_mixers,
            mixer_weights=mixer_weights,
            fixed_confidence=fixed_confidence,
            anomaly_cooldown=anomaly_cooldown,
            forecast_offset=forecast_offset,
            simple_ts_bounds=simple_ts_bounds,
            time_format=time_format,
            force_ts_infer=force_ts_infer,
            return_embedding=return_embedding,
        )

        return pred_args

    def to_dict(self, encode_json=False) -> Dict[str, Json]:
        """
        Creates a python dictionary from the ``PredictionArguments`` object

        :returns: A python dictionary
        """
        return _asdict(self, encode_json=encode_json)



import torch
import torch.nn as nn
import numpy as np
from lightwood.helpers.torch import LightwoodAutocast


class DecoderRNNNumerical(nn.Module):
    def __init__(self, hidden_size, output_size):
        super(DecoderRNNNumerical, self).__init__()
        self.hidden_size = hidden_size
        self.in_activation = nn.Sigmoid()
        self.dropout = nn.Dropout(0.2)
        self.gru = nn.GRU(output_size, hidden_size, batch_first=True)
        self.out = nn.Linear(hidden_size, output_size)

    def forward(self, input, hidden):
        with LightwoodAutocast():
            output = self.in_activation(input.float())
            output, hidden = self.gru(output, hidden)
            output = self.dropout(output)
            output = self.out(output)
        return output, hidden

    def init_hidden(self, device, batch_size=1):
        return torch.zeros(1, batch_size, self.hidden_size, device=device)

    def decode(self, data, initial_tensor, criterion, device, hidden_state=None, sos=0):
        """This method decodes an input unrolled through time, given an initial hidden state"""
        if isinstance(data, tuple):
            # when using transformer encoder, data contains a sequence length tensor
            data, len_data = data

        if initial_tensor.shape[1] > 1:
            # if tensor is a sequence (like the one yielded by the transformer),
            # we select only the last timestep for decoding
            initial_tensor = initial_tensor[:, -1:, :]

        loss = 0
        next_tensor = torch.full_like(initial_tensor, sos, dtype=torch.float32).to(device)
        tensor_target = torch.cat([next_tensor, data], dim=1)  # add SOS token at t=0 to true input
        if hidden_state is None:
            hidden_state = self.init_hidden(device, data.shape[0])

        for tensor_i in range(data.shape[1] - 1):
            rand = np.random.randint(2)
            # teach from forward as well as from known tensor alternatively
            if rand == 1:
                next_tensor, hidden_state = self.forward(tensor_target[:, tensor_i, :].unsqueeze(dim=1), hidden_state)
            else:
                next_tensor, hidden_state = self.forward(next_tensor.detach(), hidden_state)

            loss += criterion(next_tensor, tensor_target[:, tensor_i + 1, :].unsqueeze(dim=1))

        return next_tensor, hidden_state, loss


class EncoderRNNNumerical(nn.Module):
    def __init__(self, input_size, hidden_size):
        super(EncoderRNNNumerical, self).__init__()
        self.hidden_size = hidden_size
        self.dropout = nn.Dropout(0.2)
        self.gru = nn.GRU(input_size, hidden_size, batch_first=True)
        self.out = nn.Linear(hidden_size, input_size)

    def forward(self, input, hidden):
        with LightwoodAutocast():
            output, hidden = self.gru(input, hidden)
            output = self.dropout(output)
            output = self.out(output)
        return output, hidden

    def init_hidden(self, device, batch_size=1):
        return torch.zeros(1, batch_size, self.hidden_size, device=device)

    def bptt(self, data, criterion, device):
        """This method encodes an input unrolled through time"""
        loss = 0
        hidden_state = self.init_hidden(device, batch_size=data.shape[0])
        next_tensor = data[:, 0, :].unsqueeze(dim=1)  # initial input

        for tensor_i in range(data.shape[1] - 1):
            rand = np.random.randint(2)
            # teach from forward as well as from known tensor alternatively
            if rand == 1:
                next_tensor, hidden_state = self.forward(data[:, tensor_i, :].unsqueeze(dim=1), hidden_state)
            else:
                next_tensor, hidden_state = self.forward(next_tensor.detach(), hidden_state)

            loss += criterion(next_tensor, data[:, tensor_i + 1, :].unsqueeze(dim=1))

        return next_tensor, hidden_state, loss



import os
from typing import Union
import dill
import pandas as pd
from lightwood.api.types import JsonAI, ProblemDefinition
from dataprep_ml.insights import statistical_analysis
from type_infer.api import infer_types
from lightwood.api.predictor import PredictorInterface
from lightwood.api.json_ai import generate_json_ai
from lightwood.helpers.codegen import code_from_json_ai as _code_from_json_ai, _module_from_code, _predictor_from_code
import sys
import gc
import time
from lightwood.helpers.log import log
from shutil import copyfile


def load_custom_module(file_path: str):
    from lightwood import __version__ as lightwood_version
    modules_dir = os.path.join(os.path.expanduser('~/lightwood_modules'), lightwood_version.replace('.', '_'))
    custom_module_path = os.path.join(modules_dir, os.path.split(file_path)[-1])
    if not os.path.exists(modules_dir):
        os.makedirs(modules_dir)
    if os.path.exists(custom_module_path):
        os.remove(custom_module_path)

    copyfile(file_path, custom_module_path)


def predictor_from_problem(df: pd.DataFrame, problem_definition: Union[ProblemDefinition, dict]) -> PredictorInterface:
    """
    Creates a ready-to-train ``Predictor`` object from some raw data and a ``ProblemDefinition``. Do not use this if you want to edit the JsonAI first. Usually you'd want to next train this predictor by calling the ``learn`` method on the same dataframe used to create it.

    :param df: The raw data
    :param problem_definition: The manual specifications for your predictive problem

    :returns: A lightwood ``Predictor`` object
    """ # noqa
    if not isinstance(problem_definition, ProblemDefinition):
        problem_definition = ProblemDefinition.from_dict(problem_definition)

    if problem_definition.ignore_features:
        log.info(f'Dropping features: {problem_definition.ignore_features}')
        df = df.drop(columns=problem_definition.ignore_features)

    predictor_class_str = code_from_problem(df, problem_definition)
    return predictor_from_code(predictor_class_str)


def json_ai_from_problem(df: pd.DataFrame, problem_definition: Union[ProblemDefinition, dict]) -> JsonAI:
    """
    Creates a JsonAI from your raw data and problem definition. Usually you would use this when you want to subsequently edit the JsonAI, the easiest way to do this is to unload it to a dictionary via `to_dict`, modify it, and then create a new object from it using `lightwood.JsonAI.from_dict`. It's usually better to generate the JsonAI using this function rather than writing it from scratch.

    :param df: The raw data
    :param problem_definition: The manual specifications for your predictive problem

    :returns: A ``JsonAI`` object generated based on your data and problem specifications
    """ # noqa
    if not isinstance(problem_definition, ProblemDefinition):
        problem_definition = ProblemDefinition.from_dict(problem_definition)

    started = time.time()

    if problem_definition.ignore_features:
        log.info(f'Dropping features: {problem_definition.ignore_features}')
        df = df.drop(columns=problem_definition.ignore_features)

    type_information = infer_types(df, config={'engine': 'rule_based', 'pct_invalid': problem_definition.pct_invalid})
    stats = statistical_analysis(
        df, type_information.dtypes, problem_definition.to_dict(), type_information.identifiers)

    duration = time.time() - started
    if problem_definition.time_aim is not None:
        problem_definition.time_aim -= duration
        if problem_definition.time_aim < 10:
            problem_definition.time_aim = 10

    # Assume that the stuff besides encoder and mixers takes about as long as analyzing did... bad, but let's see
    if problem_definition.expected_additional_time is None:
        problem_definition.expected_additional_time = duration
    json_ai = generate_json_ai(
        type_information=type_information, statistical_analysis=stats,
        problem_definition=problem_definition)

    return json_ai


def code_from_json_ai(json_ai: JsonAI) -> str:
    """
    Autogenerates custom code based on the details you specified inside your JsonAI.

    :param json_ai: A ``JsonAI`` object

    :returns: Code (text) generate based on the ``JsonAI`` you created
    """
    return _code_from_json_ai(json_ai)


def predictor_from_code(code: str) -> PredictorInterface:
    """
    :param code: The ``Predictor``'s code in text form

    :returns: A lightwood ``Predictor`` object
    """
    return _predictor_from_code(code)


def code_from_problem(df: pd.DataFrame, problem_definition: Union[ProblemDefinition, dict]) -> str:
    """
    :param df: The raw data
    :param problem_definition: The manual specifications for your predictive problem

    :returns: The text code generated based on your data and problem specifications
    """
    if not isinstance(problem_definition, ProblemDefinition):
        problem_definition = ProblemDefinition.from_dict(problem_definition)

    if problem_definition.ignore_features:
        log.info(f'Dropping features: {problem_definition.ignore_features}')
        df = df.drop(columns=problem_definition.ignore_features)

    json_ai = json_ai_from_problem(df, problem_definition)
    predictor_code = code_from_json_ai(json_ai)
    return predictor_code


def predictor_from_state(state_file: str, code: str = None) -> PredictorInterface:
    """
    :param state_file: The file containing the pickle resulting from calling ``save`` on a ``Predictor`` object
    :param code: The ``Predictor``'s code in text form

    :returns: A lightwood ``Predictor`` object
    """
    try:
        module_name = None
        with open(state_file, 'rb') as fp:
            predictor = dill.load(fp)
    except Exception as e:
        module_name = str(e).lstrip("No module named '").split("'")[0]
        if code is None:
            raise Exception(
                'Provide code when loading a predictor from outside the scope/script it was created in!')

    if module_name is not None:
        try:
            del sys.modules[module_name]
        except Exception:
            pass
        gc.collect()
        _module_from_code(code, module_name)
        with open(state_file, 'rb') as fp:
            predictor = dill.load(fp)

    return predictor


def predictor_from_json_ai(json_ai: JsonAI) -> PredictorInterface:
    """
    Creates a ready-to-train ``Predictor`` object based on the details you specified inside your JsonAI.

    :param json_ai: A ``JsonAI`` object

    :returns: A lightwood ``Predictor`` object
    """  # noqa
    code = code_from_json_ai(json_ai)
    predictor = predictor_from_code(code)
    return predictor



import torch
import torch.nn as nn
from lightwood.helpers.device import get_device_from_name
from lightwood.helpers.torch import LightwoodAutocast

from lightwood.helpers.log import log

try:
    import torchvision.models as models
except ModuleNotFoundError:
    log.info("No torchvision detected, image helpers not supported.")


class ChannelPoolAdaptiveAvg1d(torch.nn.AdaptiveAvgPool1d):
    """
    Custom override of `torch.nn.AdaptiveAvgPool1d` to use `LightwoodAutocast()` and handle dimensions in the way we need to.
    """  # noqa
    def forward(self, input):
        with LightwoodAutocast():
            n, c, _, _ = input.size()
            input = input.view(n, c, 1).permute(0, 2, 1)
            pooled = torch.nn.functional.adaptive_avg_pool1d(input, self.output_size)
            _, _, c = pooled.size()
            pooled = pooled.permute(0, 2, 1)
            return pooled.view(n, c)


class Img2Vec(nn.Module):
    """ 
    Img2Vec is a ``torch.nn.module`` that returns image embeddings.
    
    For this, it uses a pretrained `torchvision.torch.resnext50_32x4d`, with its final fully connected layer removed.
    
    Output is a `self.output_size`-dimensioned vector, generated by taking the output of the Resnext's last convolutional layer and performing an adaptive channel pool average. 
    """  # noqa
    def __init__(self, device=''):
        super(Img2Vec, self).__init__()

        self.device = get_device_from_name(device)

        self.output_size = 512
        self.model = torch.nn.Sequential(*list(models.resnext50_32x4d(pretrained=True).children())[: -1],
                                         ChannelPoolAdaptiveAvg1d(output_size=self.output_size))
        self.model = self.model.to(self.device)

    def to(self, device, available_devices=1):
        self.device = device
        self.model = self.model.to(self.device)
        return self

    def forward(self, image, batch=True):
        with LightwoodAutocast():
            embedding = self.model(image.to(self.device))

            if batch:
                return embedding
            return embedding[0, :]



from lightwood.mixer.base import BaseMixer
from lightwood.mixer.unit import Unit
from lightwood.mixer.neural import Neural
from lightwood.mixer.neural_ts import NeuralTs
from lightwood.mixer.xgboost import XGBoostMixer
from lightwood.mixer.xgboost_array import XGBoostArrayMixer
from lightwood.mixer.random_forest import RandomForest
from lightwood.mixer.sktime import SkTime
from lightwood.mixer.arima import ARIMAMixer
from lightwood.mixer.ets import ETSMixer
from lightwood.mixer.regression import Regression

try:
    from lightwood.mixer.qclassic import QClassic
except Exception:
    QClassic = None

try:
    from lightwood.mixer.nhits import NHitsMixer
except Exception:
    NHitsMixer = None

try:
    from lightwood.mixer.prophet import ProphetMixer
except Exception:
    ProphetMixer = None

try:
    from lightwood.mixer.gluonts import GluonTSMixer
except Exception:
    GluonTSMixer = None

try:
    from lightwood.mixer.lightgbm import LightGBM
    from lightwood.mixer.lightgbm_array import LightGBMArray
except Exception:
    LightGBM = None
    LightGBMArray = None

try:
    from lightwood.mixer.tabtransformer import TabTransformerMixer
except Exception:
    TabTransformerMixer = None

__all__ = ['BaseMixer', 'Neural', 'NeuralTs', 'LightGBM', 'RandomForest', 'LightGBMArray', 'Unit', 'Regression',
           'SkTime', 'QClassic', 'ProphetMixer', 'ETSMixer', 'ARIMAMixer', 'NHitsMixer', 'GluonTSMixer', 'XGBoostMixer',
           'TabTransformerMixer', 'XGBoostArrayMixer']
